from typing import Optional

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__EXECUTION__"


class ExecutionSettings(FidesSettings):
    """Configuration settings for DSR execution."""

    masking_strict: bool = Field(
        default=True,
        description="If set to True, only use UPDATE requests to mask data. If False, Fides will use any defined DELETE or GDPR DELETE endpoints to remove PII, which may extend beyond the specific data categories that configured in your execution policy.",
    )
    privacy_request_delay_timeout: int = Field(
        default=3600,
        description="The amount of time to wait for actions which delay privacy requests (e.g., pre- and post-processing webhooks).",
    )
    require_manual_request_approval: bool = Field(
        default=False,
        description="Whether access and erasure privacy requests require explicit approval to execute. Consent privacy requests are always auto-approved.",
    )
    subject_identity_verification_required: bool = Field(
        default=False,
        description="Whether privacy requests require user identity verification.",
    )
    disable_consent_identity_verification: Optional[bool] = Field(
        default=None,
        description="Allows selective disabling of identity verification specifically for consent requests. Identity verification for consent requests will be enabled if subject_identity_verification_required is set to true and this setting is empty or false.",
    )
    task_retry_backoff: int = Field(
        default=1,
        description="The backoff factor for retries, to space out repeated retries.",
    )
    task_retry_count: int = Field(
        default=0, description="The number of times a failed request will be retried."
    )
    task_retry_delay: int = Field(
        default=1, description="The delays between retries in seconds."
    )
    allow_custom_privacy_request_field_collection: bool = Field(
        default=False,
        description="Allows the collection of custom privacy request fields from incoming privacy requests.",
    )
    allow_custom_privacy_request_fields_in_request_execution: bool = Field(
        default=False,
        description="Allows custom privacy request fields to be used in request execution.",
    )
    request_task_ttl: int = Field(
        default=604800,
        description="The number of seconds a request task should live.",
    )
    state_polling_interval: int = Field(
        default=30,
        description="Seconds between polling for Privacy Requests that should change state",
    )
    interrupted_task_requeue_interval: int = Field(
        default=300,
        description="Seconds between polling for interrupted tasks to requeue",
    )
    use_dsr_3_0: bool = Field(
        default=False,
        description="Temporary flag to switch to using DSR 3.0 to process your tasks.",
    )
    fuzzy_search_enabled: bool = Field(
        default=True,
        description="Whether fuzzy search is enabled for privacy request lookups.",
    )
    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)



"""Contains all of the ungrouped CLI commands for fides."""

from datetime import datetime, timezone

import rich_click as click
import yaml

import fides
from fides.cli.options import (
    dry_flag,
    fides_key_argument,
    manifests_dir_argument,
    resource_type_argument,
    verbose_flag,
)
from fides.cli.utils import (
    FIDES_ASCII_ART,
    check_server,
    send_init_analytics,
    with_analytics,
    with_server_health_check,
)
from fides.common.utils import (
    echo_green,
    echo_red,
    handle_cli_response,
    pretty_echo,
    print_divider,
)
from fides.config.create import create_and_update_config_file
from fides.core import api as _api
from fides.core import audit as _audit
from fides.core import evaluate as _evaluate
from fides.core import parse as _parse
from fides.core import push as _push
from fides.core.api_helpers import get_server_resource, list_server_resources


@click.command()  # type: ignore
@click.pass_context
@resource_type_argument
@fides_key_argument
@with_analytics
def delete(ctx: click.Context, resource_type: str, fides_key: str) -> None:
    """
    Delete an object from the server.
    """
    config = ctx.obj["CONFIG"]
    handle_cli_response(
        _api.delete(
            url=config.cli.server_url,
            resource_type=resource_type,
            resource_id=fides_key,
            headers=config.user.auth_header,
        ),
        verbose=False,
    )
    echo_green(
        f"{resource_type.capitalize()} with fides_key '{fides_key}' successfully deleted."
    )


@click.command(name="get")  # type: ignore
@click.pass_context
@resource_type_argument
@fides_key_argument
@with_analytics
@with_server_health_check
def get_resource(ctx: click.Context, resource_type: str, fides_key: str) -> None:
    """
    View an object from the server.
    """
    config = ctx.obj["CONFIG"]
    resource = get_server_resource(
        url=config.cli.server_url,
        resource_type=resource_type,
        resource_key=fides_key,
        headers=config.user.auth_header,
    )
    print_divider()
    echo_green(yaml.dump({resource_type: [resource]}))


@click.command(name="ls")  # type: ignore
@click.pass_context
@resource_type_argument
@with_analytics
@with_server_health_check
@click.option(
    "--verbose", "-v", is_flag=True, help="Displays the entire object list as YAML."
)
def list_resources(ctx: click.Context, verbose: bool, resource_type: str) -> None:
    """
    View all objects of a single type from the server.
    """
    config = ctx.obj["CONFIG"]
    resources = list_server_resources(
        url=config.cli.server_url,
        resource_type=resource_type,
        headers=config.user.auth_header,
        exclude_keys=[],
    )
    print_divider()
    if verbose:
        echo_green(yaml.dump({resource_type: resources}))
    else:
        if resources:
            sorted_fides_keys = sorted(
                {resource["fides_key"] for resource in resources if resource}  # type: ignore[index]
            )
            formatted_fides_keys = "\n  ".join(sorted_fides_keys)
            echo_green(
                f"List of resources of type '{resource_type}':\n  {formatted_fides_keys}"
            )
        else:
            echo_red(f"No {resource_type.capitalize()} resources found!")


@click.command()  # type: ignore
@click.pass_context
@click.argument("fides_dir", default=".", type=click.Path(exists=True))
@click.option(
    "--opt-in", is_flag=True, help="Automatically opt-in to anonymous usage analytics."
)
def init(ctx: click.Context, fides_dir: str, opt_in: bool) -> None:
    """
    Initializes a Fides instance by creating the default directory and
    configuration file if not present.
    """

    executed_at = datetime.now(timezone.utc)
    config = ctx.obj["CONFIG"]

    click.echo(FIDES_ASCII_ART)
    click.echo("Initializing fides...")

    config, config_path = create_and_update_config_file(
        config, fides_dir, opt_in=opt_in
    )

    print_divider()

    send_init_analytics(config.user.analytics_opt_out, config_path, executed_at)
    echo_green("fides initialization complete.")


@click.command()
@click.pass_context
@with_analytics
def status(ctx: click.Context) -> None:
    """
    Check Fides server availability.
    """
    config = ctx.obj["CONFIG"]
    cli_version = fides.__version__
    server_url = config.cli.server_url
    click.echo("Getting server status...")
    check_server(
        cli_version=cli_version,
        server_url=server_url,
    )


@click.command()  # type: ignore
@click.pass_context
@click.option("--port", "-p", type=int, default=8080)
def webserver(ctx: click.Context, port: int = 8080) -> None:
    """
    Start the Fides webserver.

    _Requires Redis and Postgres to be configured and running_
    """
    # This has to be here to avoid a circular dependency
    from fides.api.main import start_webserver

    start_webserver(port=port)


@click.command()
@click.pass_context
@click.option("--queues", "-q", type=str, default="")
@click.option("--exclude-queues", type=str, default="")
@with_analytics
def worker(ctx: click.Context, queues: str = "", exclude_queues: str = "") -> None:
    """
    Start a Celery worker for the Fides webserver.
    """
    # This has to be here to avoid a circular dependency
    from fides.api.worker import start_worker

    start_worker(queues, exclude_queues)


@click.command()  # type: ignore
@click.pass_context
@dry_flag
@click.option(
    "--diff",
    is_flag=True,
    help="Print any diffs between the local & server objects",
)
@manifests_dir_argument
@with_analytics
@with_server_health_check
def push(ctx: click.Context, dry: bool, diff: bool, manifests_dir: str) -> None:
    """
    Parse local manifest files and upload them to the server.
    """
    config = ctx.obj["CONFIG"]
    taxonomy = _parse.parse(manifests_dir)
    _push.push(
        url=config.cli.server_url,
        taxonomy=taxonomy,
        headers=config.user.auth_header,
        dry=dry,
        diff=diff,
    )


@click.command()  # type: ignore
@click.pass_context
@manifests_dir_argument
@click.option(
    "--fides-key",
    "-k",
    help="The fides_key of a specific policy to evaluate.",
    default="",
)
@click.option(
    "-m",
    "--message",
    help="Describe the context of this evaluation.",
)
@click.option(
    "-a",
    "--audit",
    is_flag=True,
    help="Validate that the objects in this evaluation produce a valid data map.",
)
@click.option(
    "--dry",
    is_flag=True,
    help="Do not upload objects or results to the Fides webserver.",
)
@with_analytics
def evaluate(
    ctx: click.Context,
    manifests_dir: str,
    fides_key: str,
    message: str,
    audit: bool,
    dry: bool,
) -> None:
    """
    Evaluate System-level Privacy Declarations against Organization-level Policy Rules.
    """

    config = ctx.obj["CONFIG"]

    if config.cli.local_mode:
        dry = True
    else:
        taxonomy = _parse.parse(manifests_dir)
        _push.push(
            url=config.cli.server_url,
            taxonomy=taxonomy,
            headers=config.user.auth_header,
            dry=dry,
        )

    _evaluate.evaluate(
        url=config.cli.server_url,
        headers=config.user.auth_header,
        manifests_dir=manifests_dir,
        policy_fides_key=fides_key,
        message=message,
        local=config.cli.local_mode,
        dry=dry,
    )

    if audit:
        taxonomy = _parse.parse(manifests_dir)
        print_divider()
        pretty_echo("Auditing Organization Resource Compliance")
        _audit.audit_organizations(
            url=config.cli.server_url,
            headers=config.user.auth_header,
            include_keys=[
                organization.fides_key for organization in taxonomy.organization
            ],
        )
        print_divider()
        pretty_echo("Auditing System Resource Compliance")
        _audit.audit_systems(
            url=config.cli.server_url,
            headers=config.user.auth_header,
            include_keys=[system.fides_key for system in taxonomy.system or []],
        )


@click.command()
@click.pass_context
@manifests_dir_argument
@verbose_flag
@with_analytics
def parse(ctx: click.Context, manifests_dir: str, verbose: bool = False) -> None:
    """
    Parse all Fides objects located in the supplied directory.
    """
    taxonomy = _parse.parse(manifests_dir=manifests_dir)
    if verbose:
        pretty_echo(taxonomy.model_dump(mode="json"), color="green")



"""
- add page column to asset table and
- change `Browser Request` values to `Browser request` StagedResource.resource_type column
- change `Browser Request` values to `Browser request` Asset.asset_type column

Revision ID: 1d2f04ac19f2
Revises: 7c3fbee90c78
Create Date: 2025-03-17 19:36:43.016383

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "1d2f04ac19f2"
down_revision = "7c3fbee90c78"
branch_labels = None
depends_on = None


def upgrade():
    # adds page column to asset table
    op.add_column(
        "asset",
        sa.Column("page", sa.ARRAY(sa.String()), server_default="{}", nullable=False),
    )

    # changes `Browser Request` values to `Browser request` in the two tables where it shows up`
    op.execute(
        "UPDATE stagedresource SET resource_type = 'Browser request' WHERE resource_type = 'Browser Request'"
    )
    op.execute(
        "UPDATE asset SET asset_type = 'Browser request' WHERE asset_type = 'Browser Request'"
    )


def downgrade():
    # drops page column from asset table
    op.drop_column("asset", "page")

    # changes `Browser request` values back to `Browser Request` in the two tables where it shows up
    op.execute(
        "UPDATE stagedresource SET resource_type = 'Browser Request' WHERE resource_type = 'Browser request'"
    )
    op.execute(
        "UPDATE asset SET asset_type = 'Browser Request' WHERE asset_type = 'Browser request'"
    )



"""add_compass_system_sync

Revision ID: 842790fa918a
Revises: 2af81f2b1a6f
Create Date: 2023-11-07 16:22:01.746436

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "842790fa918a"
down_revision = "2af81f2b1a6f"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "system_compass_sync",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("sync_started_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("sync_completed_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("updated_systems", postgresql.ARRAY(sa.String()), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_system_compass_sync_id"), "system_compass_sync", ["id"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_system_compass_sync_id"), table_name="system_compass_sync")
    op.drop_table("system_compass_sync")



from __future__ import annotations

import json
from datetime import datetime
from typing import Any, Optional

from sqlalchemy import ARRAY, Column, ForeignKey, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session

from fides.api.cryptography.cryptographic_util import (
    generate_salt,
    generate_secure_random_string,
    hash_credential_with_salt,
)
from fides.api.cryptography.schemas.jwt import (
    JWE_ISSUED_AT,
    JWE_PAYLOAD_CLIENT_ID,
    JWE_PAYLOAD_CONNECTIONS,
    JWE_PAYLOAD_ROLES,
    JWE_PAYLOAD_SCOPES,
    JWE_PAYLOAD_SYSTEMS,
)
from fides.api.db.base_class import Base
from fides.api.models.fides_user import FidesUser
from fides.api.oauth.jwt import generate_jwe
from fides.config import FidesConfig

DEFAULT_SCOPES: list[str] = []
DEFAULT_ROLES: list[str] = []
DEFAULT_SYSTEMS: list[str] = []
DEFAULT_CONNECTIONS: list[str] = []


class ClientDetail(Base):
    """The persisted details about a client in the system"""

    @declared_attr
    def __tablename__(self) -> str:
        return "client"

    hashed_secret = Column(String, nullable=False)
    salt = Column(String, nullable=False)
    scopes = Column(ARRAY(String), nullable=False, server_default="{}", default=dict)
    roles = Column(ARRAY(String), nullable=False, server_default="{}", default=dict)
    systems = Column(ARRAY(String), nullable=False, server_default="{}", default=dict)
    connections = Column(
        ARRAY(String), nullable=False, server_default="{}", default=dict
    )
    fides_key = Column(String, index=True, unique=True, nullable=True)
    user_id = Column(
        String, ForeignKey(FidesUser.id_field_path), nullable=True, unique=True
    )

    @classmethod
    def create_client_and_secret(
        cls,
        db: Session,
        client_id_byte_length: int,
        client_secret_byte_length: int,
        *,
        scopes: list[str] | None = None,
        fides_key: Optional[str] = None,
        user_id: Optional[str] = None,
        encoding: str = "UTF-8",
        roles: list[str] | None = None,
        systems: list[str] | None = None,
        connections: list[str] | None = None,
        in_memory: bool | None = False,
    ) -> tuple["ClientDetail", str]:
        """Creates a ClientDetail and returns that along with the unhashed secret
        so it can be returned to the user on create
        """

        client_id = generate_secure_random_string(client_id_byte_length)
        secret = generate_secure_random_string(client_secret_byte_length)

        if not scopes:
            scopes = DEFAULT_SCOPES

        if not roles:
            roles = DEFAULT_ROLES

        if not systems:
            systems = DEFAULT_SYSTEMS

        if not connections:
            connections = DEFAULT_CONNECTIONS

        salt = generate_salt()
        hashed_secret = hash_credential_with_salt(
            secret.encode(encoding),
            salt.encode(encoding),
        )

        data = {
            "id": client_id,
            "salt": salt,
            "hashed_secret": hashed_secret,
            "scopes": scopes,
            "fides_key": fides_key,
            "user_id": user_id,
            "roles": roles,
            "systems": systems,
            "connections": connections,
        }

        if in_memory:
            client = ClientDetail(
                **data
            )  # For creating a temporary ClientDetail for invalid user login flow
        else:
            client = super().create(
                db,
                data=data,
            )
        return client, secret  # type: ignore

    @classmethod
    def get(  # type: ignore
        cls,
        db: Session,
        *,
        object_id: Any,
        config: FidesConfig,
        scopes: list[str] = [],
        roles: list[str] = [],
    ) -> ClientDetail | None:
        """Fetch a database record via a client_id"""
        if object_id == config.security.oauth_root_client_id:
            return _get_root_client_detail(config, scopes=scopes, roles=roles)
        return super().get(db, object_id=object_id)

    def create_access_code_jwe(self, encryption_key: str) -> str:
        """Generates a JWE from the client detail provided"""
        payload = {
            # client id may not be necessary
            JWE_PAYLOAD_CLIENT_ID: self.id,
            JWE_PAYLOAD_SCOPES: self.scopes,
            JWE_ISSUED_AT: datetime.now().isoformat(),
            JWE_PAYLOAD_ROLES: self.roles,
            JWE_PAYLOAD_SYSTEMS: self.systems,
            JWE_PAYLOAD_CONNECTIONS: self.connections,
        }
        return generate_jwe(json.dumps(payload), encryption_key)

    def credentials_valid(self, provided_secret: str, encoding: str = "UTF-8") -> bool:
        """Verifies that the provided secret is correct."""
        provided_secret_hash = hash_credential_with_salt(
            provided_secret.encode(encoding),
            self.salt.encode(encoding),
        )

        return provided_secret_hash == self.hashed_secret


def _get_root_client_detail(
    config: FidesConfig,
    scopes: list[str],
    roles: list[str],
    encoding: str = "UTF-8",
) -> ClientDetail | None:
    """
    Return a root ClientDetail
    """
    if not config.security.oauth_root_client_secret_hash:
        raise ValueError("A root client hash is required")

    if scopes or roles:
        return ClientDetail(
            id=config.security.oauth_root_client_id,
            hashed_secret=config.security.oauth_root_client_secret_hash[0],
            salt=config.security.oauth_root_client_secret_hash[1].decode(encoding),
            scopes=scopes,
            roles=roles,
            systems=[],
            connections=[],
        )

    return ClientDetail(
        id=config.security.oauth_root_client_id,
        hashed_secret=config.security.oauth_root_client_secret_hash[0],
        salt=config.security.oauth_root_client_secret_hash[1].decode(encoding),
        scopes=DEFAULT_SCOPES,
        roles=DEFAULT_ROLES,
        systems=DEFAULT_SYSTEMS,
        connections=DEFAULT_CONNECTIONS,
    )



# type: ignore
"""
Contains all of the factory functions to generate generic CRUD endpoints.

Mostly used for `ctl`-related objects.
"""
from typing import Dict, List

from fastapi import Depends, Response, Security, status
from fideslang import FidesModelType
from sqlalchemy.ext.asyncio import AsyncSession

from fides.api.db.crud import (
    create_resource,
    delete_resource,
    get_resource_with_custom_fields,
    list_resource,
    update_resource,
    upsert_resources,
)
from fides.api.db.ctl_session import get_async_db
from fides.api.models.sql_models import ModelWithDefaultField, sql_model_map
from fides.api.oauth.utils import verify_oauth_client_prod
from fides.api.util import errors
from fides.api.util.api_router import APIRouter
from fides.api.util.endpoint_utils import (
    API_PREFIX,
    CLI_SCOPE_PREFIX_MAPPING,
    forbid_if_default,
    forbid_if_editing_any_is_default,
    forbid_if_editing_is_default,
)
from fides.common.api.scope_registry import CREATE, DELETE, READ, UPDATE


def generic_router_factory(fides_model: FidesModelType, model_type: str) -> APIRouter:
    """
    Compose all of the individual route factories into a single coherent Router.
    """

    object_router = APIRouter()

    create_router = create_router_factory(
        fides_model=fides_model, model_type=model_type
    )
    list_router = list_router_factory(fides_model=fides_model, model_type=model_type)
    get_router = get_router_factory(fides_model=fides_model, model_type=model_type)
    delete_router = delete_router_factory(
        fides_model=fides_model, model_type=model_type
    )
    update_router = update_router_factory(
        fides_model=fides_model, model_type=model_type
    )
    upsert_router = upsert_router_factory(
        fides_model=fides_model, model_type=model_type
    )

    object_router.include_router(create_router)
    object_router.include_router(list_router)
    object_router.include_router(get_router)
    object_router.include_router(delete_router)
    object_router.include_router(update_router)
    object_router.include_router(upsert_router)

    return object_router


def create_router_factory(fides_model: FidesModelType, model_type: str) -> APIRouter:
    """Return a configured version of a generic 'Create' route."""

    router = APIRouter(prefix=f"{API_PREFIX}/{model_type}", tags=[fides_model.__name__])

    @router.post(
        name="Create",
        path="/",
        response_model=fides_model,
        status_code=status.HTTP_201_CREATED,
        dependencies=[
            Security(
                verify_oauth_client_prod,
                scopes=[f"{CLI_SCOPE_PREFIX_MAPPING[model_type]}:{CREATE}"],
            )
        ],
        responses={
            status.HTTP_403_FORBIDDEN: {
                "content": {
                    "application/json": {
                        "example": {
                            "detail": {
                                "error": "user does not have permission to modify this resource",
                                "resource_type": model_type,
                                "fides_key": "example.key",
                            }
                        }
                    }
                }
            },
        },
    )
    async def create(
        resource: fides_model,
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """
        Create a resource.

        Payloads with an is_default field can only be set to False,
        will return a `403 Forbidden`.
        """
        sql_model = sql_model_map[model_type]
        if isinstance(resource, ModelWithDefaultField) and resource.is_default:
            raise errors.ForbiddenIsDefaultTaxonomyError(
                model_type, resource.fides_key, action="create"
            )

        return await create_resource(sql_model, resource.model_dump(mode="json"), db)

    return router


def list_router_factory(fides_model: FidesModelType, model_type: str) -> APIRouter:
    """Return a configured version of a generic 'List' router."""

    router = APIRouter(prefix=f"{API_PREFIX}/{model_type}", tags=[fides_model.__name__])

    @router.get(
        path="/",
        dependencies=[
            Security(
                verify_oauth_client_prod,
                scopes=[f"{CLI_SCOPE_PREFIX_MAPPING[model_type]}:{READ}"],
            )
        ],
        response_model=List[fides_model],
        name="List",
    )
    async def ls(  # pylint: disable=invalid-name
        db: AsyncSession = Depends(get_async_db),
    ) -> List:
        """Get a list of all of the resources of this type."""
        sql_model = sql_model_map[model_type]
        return await list_resource(sql_model, db)

    return router


def get_router_factory(fides_model: FidesModelType, model_type: str) -> APIRouter:
    """Return a configured version of a generic 'Get' endpoint."""

    router = APIRouter(prefix=f"{API_PREFIX}/{model_type}", tags=[fides_model.__name__])

    @router.get(
        path="/{fides_key}",
        dependencies=[
            Security(
                verify_oauth_client_prod,
                scopes=[f"{CLI_SCOPE_PREFIX_MAPPING[model_type]}:{READ}"],
            )
        ],
        response_model=fides_model,
    )
    async def get(
        fides_key: str,
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """Get a resource by its fides_key."""
        sql_model = sql_model_map[model_type]
        return await get_resource_with_custom_fields(sql_model, fides_key, db)

    return router


def update_router_factory(fides_model: FidesModelType, model_type: str) -> APIRouter:
    """Return a configured version of a generic 'Update' route."""

    router = APIRouter(prefix=f"{API_PREFIX}/{model_type}", tags=[fides_model.__name__])

    @router.put(
        path="/",
        response_model=fides_model,
        dependencies=[
            Security(
                verify_oauth_client_prod,
                scopes=[f"{CLI_SCOPE_PREFIX_MAPPING[model_type]}:{UPDATE}"],
            )
        ],
        responses={
            status.HTTP_403_FORBIDDEN: {
                "content": {
                    "application/json": {
                        "example": {
                            "detail": {
                                "error": "user does not have permission to modify this resource",
                                "resource_type": model_type,
                                "fides_key": "example.key",
                            }
                        }
                    }
                }
            },
        },
    )
    async def update(
        resource: fides_model,
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """
        Update a resource by its fides_key.

        The `is_default` field cannot be updated and will respond
        with a `403 Forbidden` if attempted.
        """
        sql_model = sql_model_map[model_type]
        await forbid_if_editing_is_default(sql_model, resource.fides_key, resource, db)
        return await update_resource(sql_model, resource.model_dump(mode="json"), db)

    return router


def upsert_router_factory(fides_model: FidesModelType, model_type: str) -> APIRouter:
    """Return a configured version of a generic 'Upsert' endpoint."""

    router = APIRouter(prefix=f"{API_PREFIX}/{model_type}", tags=[fides_model.__name__])

    @router.post(
        path="/upsert",
        dependencies=[
            Security(
                verify_oauth_client_prod,
                scopes=[
                    f"{CLI_SCOPE_PREFIX_MAPPING[model_type]}:{CREATE}",
                    f"{CLI_SCOPE_PREFIX_MAPPING[model_type]}:{UPDATE}",
                ],
            )
        ],
        responses={
            status.HTTP_200_OK: {
                "content": {
                    "application/json": {
                        "example": {
                            "message": f"Upserted 3 {model_type}(s)",
                            "inserted": 0,
                            "updated": 3,
                        }
                    }
                }
            },
            status.HTTP_201_CREATED: {
                "content": {
                    "application/json": {
                        "example": {
                            "message": f"Upserted 3 {model_type}(s)",
                            "inserted": 1,
                            "updated": 2,
                        }
                    }
                }
            },
            status.HTTP_403_FORBIDDEN: {
                "content": {
                    "application/json": {
                        "example": {
                            "detail": {
                                "error": "user does not have permission to modify this resource",
                                "resource_type": "DataCategory",
                                "fides_key": "example.key",
                            }
                        }
                    }
                }
            },
        },
    )
    async def upsert(
        resources: List[fides_model],
        response: Response,
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """
        For any resource in `resources` that already exists in the database,
        update the resource by its `fides_key`. Otherwise, create a new resource.

        Responds with a `201 Created` if even a single resource in `resources`
        did not previously exist. Otherwise, responds with a `200 OK`.

        The `is_default` field cannot be updated and will respond
        with a `403 Forbidden` if attempted.
        """

        sql_model = sql_model_map[model_type]
        resource_dicts = [resource.model_dump(mode="json") for resource in resources]
        await forbid_if_editing_any_is_default(sql_model, resource_dicts, db)
        result = await upsert_resources(sql_model, resource_dicts, db)
        response.status_code = (
            status.HTTP_201_CREATED if result[0] > 0 else response.status_code
        )

        return {
            "message": f"Upserted {len(resources)} {sql_model.__name__}(s)",
            "inserted": result[0],
            "updated": result[1],
        }

    return router


def delete_router_factory(fides_model: FidesModelType, model_type: str) -> APIRouter:
    """Return a configured version of a generic 'Delete' route."""

    router = APIRouter(prefix=f"{API_PREFIX}/{model_type}", tags=[fides_model.__name__])

    @router.delete(
        path="/{fides_key}",
        dependencies=[
            Security(
                verify_oauth_client_prod,
                scopes=[f"{CLI_SCOPE_PREFIX_MAPPING[model_type]}:{DELETE}"],
            )
        ],
        responses={
            status.HTTP_403_FORBIDDEN: {
                "content": {
                    "application/json": {
                        "example": {
                            "detail": {
                                "error": "user does not have permission to modify this resource",
                                "resource_type": model_type,
                                "fides_key": "example.key",
                            }
                        }
                    }
                }
            },
        },
    )
    async def delete(
        fides_key: str,
        db: AsyncSession = Depends(get_async_db),
    ) -> Dict:
        """
        Delete a resource by its fides_key.

        Resources that have `is_default=True` cannot be deleted and will respond
        with a `403 Forbidden`.
        """
        sql_model = sql_model_map[model_type]
        await forbid_if_default(sql_model, fides_key, db)
        deleted_resource = await delete_resource(sql_model, fides_key, db)
        # Convert the resource to a dict explicitly for the response
        deleted_resource_dict = fides_model.model_validate(deleted_resource).model_dump(
            mode="json"
        )
        return {
            "message": "resource deleted",
            "resource": deleted_resource_dict,
        }

    return router



"""add property table

Revision ID: 26d5976531d6
Revises: 68cb26f3492d
Create Date: 2024-02-20 18:44:57.913009

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "26d5976531d6"
down_revision = "68cb26f3492d"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "plus_property",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("type", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_index(op.f("ix_plus_property_key"), "plus_property", ["key"], unique=True)
    op.create_index(op.f("ix_plus_property_id"), "plus_property", ["id"], unique=False)


def downgrade():
    op.drop_index(op.f("ix_plus_property_id"), table_name="plus_property")
    op.drop_index(op.f("ix_plus_property_key"), table_name="plus_property")
    op.drop_table("plus_property")



"""Merge alembic heads

Revision ID: 1f61c765cd1c
Revises: 8f84fad4e00b, b72541d79f10
Create Date: 2022-12-02 17:59:08.490577

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1f61c765cd1c"
down_revision = ("8f84fad4e00b", "b72541d79f10")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from typing import Any, Dict, Optional

from fastapi import HTTPException, Query
from fastapi.params import Security
from fastapi_pagination import Page, Params, paginate
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.utils import disable_installed_extensions_check
from starlette.status import HTTP_404_NOT_FOUND

from fides.api.common_exceptions import NoSuchConnectionTypeSecretSchemaError
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.connection_configuration.connection_type_system_map import (
    ConnectionSystemTypeMap,
)
from fides.api.schemas.connection_configuration.enums.system_type import SystemType
from fides.api.schemas.policy import ActionType
from fides.api.util.api_router import APIRouter
from fides.api.util.connection_type import (
    get_connection_type_secret_schema,
    get_connection_types,
)
from fides.common.api.scope_registry import CONNECTION_TYPE_READ
from fides.common.api.v1.urn_registry import (
    CONNECTION_TYPE_SECRETS,
    CONNECTION_TYPES,
    V1_URL_PREFIX,
)

# FastAPI is saying that because cassandra is installed, we should use fastapi_pagination.ext.cassandra here,
# which is not relevant
disable_installed_extensions_check()

router = APIRouter(tags=["Connection Types"], prefix=V1_URL_PREFIX)


@router.get(
    CONNECTION_TYPES,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_TYPE_READ])],
    response_model=Page[ConnectionSystemTypeMap],
)
def get_all_connection_types(
    *,
    page: int = Query(1, ge=1, description="Page number"),
    size: int = Query(
        100, ge=1, le=100, description="Page size"
    ),  # set a larger default page size for this endpoint
    search: Optional[str] = None,
    system_type: Optional[SystemType] = None,
    consent: Optional[bool] = None,
    access: Optional[bool] = None,
    erasure: Optional[bool] = None,
) -> AbstractPage[ConnectionSystemTypeMap]:
    """
    Returns a list of connection options in Fides - includes only database and saas options here.

    Query params for types of requests supported - `consent`, `access` and `erasure` - act as filters.
    If set to `true`, only connections that support the specified type of request will be returned.
    If no filters are specified, then no filtering is performed.
    When applied together, the filters act as a union: result sets are additive.
    """
    action_types = set()
    # special-case when no action type filters are provided
    if consent is None and access is None and erasure is None:
        action_types = {ActionType.access, ActionType.erasure, ActionType.consent}
    else:
        if access:
            action_types.add(ActionType.access)
        if erasure:
            action_types.add(ActionType.erasure)
        if consent:
            action_types.add(ActionType.consent)

    params = Params(page=page, size=size)

    return paginate(
        get_connection_types(search, system_type, action_types),
        params,
    )


@router.get(
    CONNECTION_TYPE_SECRETS,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_TYPE_READ])],
)
def get_connection_type_secret_schema_route(
    *, connection_type: str
) -> Optional[Dict[str, Any]]:
    """Returns the secret fields that should be supplied to authenticate with a particular connection type

    Note that this endpoint should never return actual secrets, we return the *types* of secret fields needed
    to authenticate.
    """
    try:
        return get_connection_type_secret_schema(connection_type=connection_type)
    except NoSuchConnectionTypeSecretSchemaError:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No connection type found with name '{connection_type}'.",
        )



from sqlalchemy import Boolean, Column, Integer, String, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr

from fides.api.db.base_class import Base


class TCFPurposeOverride(Base):
    """
    Stores TCF Purpose Overrides

    Allows a customer to override Fides-wide which Purposes show up in the TCF Experience, and
    specify a global legal basis for that Purpose.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "tcf_purpose_overrides"

    purpose = Column(Integer, nullable=False, index=True)
    is_included = Column(Boolean, server_default="t", default=True)
    required_legal_basis = Column(String)

    __table_args__ = (UniqueConstraint("purpose"),)



"""Add fideslib models

Revision ID: 155fd8e51d9d
Revises: be432bd23596
Create Date: 2022-07-09 01:13:23.440193

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "155fd8e51d9d"
down_revision = "be432bd23596"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # data_categories
    op.execute("DROP SEQUENCE data_categories_id_seq CASCADE")
    op.execute(
        "ALTER TABLE data_categories DROP CONSTRAINT data_categories_pkey CASCADE"
    )
    op.alter_column(
        table_name="data_categories",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="data_categories",
        constraint_name="data_categories_pkey",
        columns=["id"],
    )

    # data_qualifiers
    op.execute("DROP SEQUENCE data_qualifiers_id_seq CASCADE")
    op.execute(
        "ALTER TABLE data_qualifiers DROP CONSTRAINT data_qualifiers_pkey CASCADE"
    )
    op.alter_column(
        table_name="data_qualifiers",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="data_qualifiers",
        constraint_name="data_qualifiers_pkey",
        columns=["id"],
    )

    # data_subjects
    op.execute("DROP SEQUENCE data_subjects_id_seq CASCADE")
    op.execute("ALTER TABLE data_subjects DROP CONSTRAINT data_subjects_pkey CASCADE")
    op.alter_column(
        table_name="data_subjects",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="data_subjects",
        constraint_name="data_subjects_pkey",
        columns=["id"],
    )

    # data_uses
    op.execute("DROP SEQUENCE data_uses_id_seq CASCADE")
    op.execute("ALTER TABLE data_uses DROP CONSTRAINT data_uses_pkey CASCADE")
    op.alter_column(
        table_name="data_uses",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="data_uses", constraint_name="data_uses_pkey", columns=["id"]
    )

    # datasets
    op.execute("DROP SEQUENCE datasets_id_seq CASCADE")
    op.execute("ALTER TABLE datasets DROP CONSTRAINT datasets_pkey CASCADE")
    op.alter_column(
        table_name="datasets",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="datasets",
        constraint_name="data_sets_pkey",
        columns=["id"],
    )

    # evaluations
    op.add_column(
        "evaluations",
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    op.add_column(
        "evaluations",
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
    )
    op.execute("DROP SEQUENCE evaluations_id_seq CASCADE")
    op.execute("ALTER TABLE evaluations DROP CONSTRAINT evaluations_pkey CASCADE")
    op.alter_column(
        table_name="evaluations",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="evaluations",
        constraint_name="evaluations_pkey",
        columns=["id"],
    )

    # organizations
    op.execute("DROP SEQUENCE organizations_id_seq CASCADE")
    op.execute("ALTER TABLE organizations DROP CONSTRAINT organizations_pkey CASCADE")
    op.alter_column(
        table_name="organizations",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="organizations", constraint_name="organizations_pkey", columns=["id"]
    )

    # policies
    op.execute("DROP SEQUENCE policies_id_seq CASCADE")
    op.execute("ALTER TABLE policies DROP CONSTRAINT policies_pkey CASCADE")
    op.alter_column(
        table_name="policies",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="policies", constraint_name="policies_pkey", columns=["id"]
    )

    # registries
    op.execute("DROP SEQUENCE registries_id_seq CASCADE")
    op.execute("ALTER TABLE registries DROP CONSTRAINT registries_pkey CASCADE")
    op.alter_column(
        table_name="registries",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="registries", constraint_name="registries_pkey", columns=["id"]
    )

    # systems
    op.execute("DROP SEQUENCE systems_id_seq CASCADE")
    op.execute("ALTER TABLE systems DROP CONSTRAINT systems_pkey CASCADE")
    op.alter_column(
        table_name="systems",
        column_name="id",
        existing_type=sa.Integer,
        type_=sa.String(255),
        nullable=False,
    )
    op.create_primary_key(
        table_name="systems", constraint_name="systems_pkey", columns=["id"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # systems
    op.execute("ALTER TABLE systems DROP CONSTRAINT systems_pkey CASCADE")
    op.execute("CREATE SEQUENCE systems_id_seq")
    op.alter_column(
        table_name="systems",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="systems", constraint_name="systems_pkey", columns=["id"]
    )

    # registries
    op.execute("ALTER TABLE registries DROP CONSTRAINT registries_pkey CASCADE")
    op.execute("CREATE SEQUENCE registries_id_seq")
    op.alter_column(
        table_name="registries",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="registries", constraint_name="registries_pkey", columns=["id"]
    )

    # policies

    op.execute("ALTER TABLE policies DROP CONSTRAINT policies_pkey CASCADE")
    op.execute("CREATE SEQUENCE policies_id_seq")
    op.alter_column(
        table_name="policies",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="policies", constraint_name="policies_pkey", columns=["id"]
    )

    # organizations
    op.execute("ALTER TABLE organizations DROP CONSTRAINT organizations_pkey CASCADE")
    op.execute("CREATE SEQUENCE policies_id_seq")
    op.alter_column(
        table_name="organizations",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="organizations", constraint_name="organizations_pkey", columns=["id"]
    )

    # evaluations
    op.execute("ALTER TABLE evaluations DROP CONSTRAINT evaluations_pkey CASCADE")
    op.execute("CREATE SEQUENCE policies_id_seq")
    op.alter_column(
        table_name="evaluations",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="evaluations", constraint_name="evaluations_pkey", columns=["id"]
    )
    op.drop_column("evaluations", "updated_at")
    op.drop_column("evaluations", "created_at")

    # datasets
    op.execute("ALTER TABLE datasets DROP CONSTRAINT datasets_pkey CASCADE")
    op.execute("CREATE SEQUENCE datasets_id_seq")
    op.alter_column(
        table_name="datasets",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="datasets", constraint_name="datasets_pkey", columns=["id"]
    )

    # data_uses
    op.execute("ALTER TABLE data_uses DROP CONSTRAINT data_uses_pkey CASCADE")
    op.execute("CREATE SEQUENCE data_uses_id_seq")
    op.alter_column(
        table_name="data_uses",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="data_uses", constraint_name="data_uses_pkey", columns=["id"]
    )

    # data_subjects
    op.execute("ALTER TABLE data_subjects DROP CONSTRAINT data_subjects_pkey CASCADE")
    op.execute("CREATE SEQUENCE data_uses_id_seq")
    op.alter_column(
        table_name="data_subjects",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="data_subjects", constraint_name="data_subjects_pkey", columns=["id"]
    )

    # data_qualifiers
    op.execute(
        "ALTER TABLE data_qualifiers DROP CONSTRAINT data_qualifiers_pkey CASCADE"
    )
    op.execute("CREATE SEQUENCE data_uses_id_seq")
    op.alter_column(
        table_name="data_qualifiers",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="data_qualifiers",
        constraint_name="data_qualifiers_pkey",
        columns=["id"],
    )

    # data_categories
    op.execute(
        "ALTER TABLE data_categories DROP CONSTRAINT data_categories_pkey CASCADE"
    )
    op.execute("CREATE SEQUENCE data_categories_id_seq")
    op.alter_column(
        table_name="data_categories",
        column_name="id",
        existing_type=sa.String(255),
        type_=sa.Integer,
        nullable=False,
    )
    op.create_primary_key(
        table_name="data_categories",
        constraint_name="data_categories_pkey",
        columns=["id"],
    )

    op.drop_index(op.f("ix_fidesuserpermissions_id"), table_name="fidesuserpermissions")
    op.drop_table("fidesuserpermissions")
    op.drop_index(op.f("ix_client_id"), table_name="client")
    op.drop_index(op.f("ix_client_fides_key"), table_name="client")
    op.drop_table("client")
    op.drop_index(op.f("ix_fidesuser_username"), table_name="fidesuser")
    op.drop_index(op.f("ix_fidesuser_id"), table_name="fidesuser")
    op.drop_table("fidesuser")
    op.drop_index(op.f("ix_auditlog_user_id"), table_name="auditlog")
    op.drop_index(op.f("ix_auditlog_privacy_request_id"), table_name="auditlog")
    op.drop_index(op.f("ix_auditlog_id"), table_name="auditlog")
    op.drop_index(op.f("ix_auditlog_action"), table_name="auditlog")
    op.drop_table("auditlog")
    # ### end Alembic commands ###



from typing import Any, List, Optional

from celery import VERSION_BANNER
from celery.apps.worker import Worker
from celery.signals import celeryd_after_setup
from loguru import logger

from fides.api.db.base import Base  # type: ignore
from fides.api.service.saas_request.override_implementations import *
from fides.api.tasks import (
    DSR_QUEUE_NAME,
    MESSAGING_QUEUE_NAME,
    PRIVACY_PREFERENCES_QUEUE_NAME,
    celery_app,
)


def start_worker(
    queues: Optional[str] = None, exclude_queues: Optional[str] = None
) -> None:
    """
    Start a Celery worker. Optionally provide a list of queues for the worker to consume,
    as a comma-separated string, or a list of queues to exclude.
    If no queues are provided, the worker will consume from all queues: the default queue,
    the messaging queue, and the privacy preferences queue.
    """

    assert (
        not queues or not exclude_queues
    ), "Cannot provide both queues and exclude_queues"

    default_queue_name = celery_app.conf.get("task_default_queue", "celery")

    all_queues = [
        default_queue_name,
        MESSAGING_QUEUE_NAME,
        PRIVACY_PREFERENCES_QUEUE_NAME,
        DSR_QUEUE_NAME,
    ]

    # Fall back to all queues if neither queues nor exclude_queues are provided.
    worker_queues = ",".join(all_queues)

    if queues:
        validate_queues(queues, all_queues)
        # If queues are provided, use them.
        worker_queues = queues

    # If excluded queues are provided, remove them from the list of all queues.
    if exclude_queues:
        validate_queues(exclude_queues, all_queues)
        excluded_queues = exclude_queues.split(",")
        worker_queues = ",".join(
            [queue for queue in all_queues if queue not in excluded_queues]
        )

    logger.info(f"Running Celery worker for queues: {worker_queues}")

    celery_app.worker_main(
        argv=[
            "--quiet",  # Disable Celery startup banner
            "worker",
            "--loglevel=info",
            "--concurrency=2",
            f"--queues={worker_queues}",
        ]
    )


def validate_queues(queues_string: str, known_queues: List[str]) -> None:
    """
    Validate that the provided queues string is a comma-separated list of known queues.
    """
    queues = queues_string.split(",")
    unknown_queues = [queue for queue in queues if queue not in known_queues]
    if unknown_queues:
        raise ValueError(f"Unknown queues: {unknown_queues}")


@celeryd_after_setup.connect
def log_celery_setup(sender: str, instance: Worker, **kwargs: Any) -> None:
    """In lieu of the Celery banner, print the connection details"""
    app = instance.app
    celery_details = {
        "hostname": instance.hostname,
        "version": VERSION_BANNER,
        "app": "{0}:{1:#x}".format(app.main or "__main__", id(app)),
        "transport": app.connection().as_uri(),
        "results": app.backend.as_uri(),
        "concurrency": str(instance.concurrency),
        "queues": "|".join(str(queue) for queue in app.amqp.queues.keys()),
    }

    logger.bind(celery_details=celery_details).info("Celery connection setup complete")


if __name__ == "__main__":  # pragma: no cover
    start_worker()



"""
Contains all of the endpoints required to validate credentials.
"""

from enum import Enum
from typing import Callable, Dict, Union

from fastapi import Response, Security, status
from pydantic import BaseModel

from fides.api.oauth.utils import verify_oauth_client_prod
from fides.api.util.api_router import APIRouter
from fides.api.util.endpoint_utils import API_PREFIX
from fides.common.api import scope_registry
from fides.connectors.models import (
    AWSConfig,
    BigQueryConfig,
    ConnectorAuthFailureException,
    ConnectorFailureException,
    OktaConfig,
)


class ValidationTarget(str, Enum):
    """
    Allowed targets for the validate endpoint
    """

    AWS = "aws"
    OKTA = "okta"
    BIGQUERY = "bigquery"


class ValidateRequest(BaseModel):
    """
    Validate endpoint request object
    """

    config: Union[AWSConfig, BigQueryConfig, OktaConfig]
    target: ValidationTarget


class ValidationStatus(str, Enum):
    """
    Validate endpoint response status
    """

    SUCCESS = "success"
    FAILURE = "failure"


class ValidateResponse(BaseModel):
    """
    Validate endpoint response object
    """

    status: ValidationStatus
    message: str


VALIDATE_ROUTER = APIRouter(tags=["Validate"], prefix=f"{API_PREFIX}/validate")


@VALIDATE_ROUTER.post(
    "/",
    dependencies=[
        Security(verify_oauth_client_prod, scopes=[scope_registry.VALIDATE_EXEC])
    ],
    status_code=status.HTTP_200_OK,
    response_model=ValidateResponse,
)
async def validate(
    validate_request_payload: ValidateRequest, response: Response
) -> ValidateResponse:
    """
    Endpoint used to validate different resource targets.
    """
    validate_function_map: Dict[ValidationTarget, Callable] = {
        ValidationTarget.AWS: validate_aws,
        ValidationTarget.BIGQUERY: validate_bigquery,
        ValidationTarget.OKTA: validate_okta,
    }
    validate_function = validate_function_map[validate_request_payload.target]

    try:
        await validate_function(validate_request_payload.config)
    except ConnectorAuthFailureException as err:
        return ValidateResponse(
            status=ValidationStatus.FAILURE,
            message=f"Authentication failed validating config. {str(err)}",
        )
    except ConnectorFailureException as err:
        return ValidateResponse(
            status=ValidationStatus.FAILURE,
            message=f"Unexpected failure validating config. {str(err)}",
        )

    return ValidateResponse(
        status=ValidationStatus.SUCCESS, message="Config validation succeeded"
    )


async def validate_aws(aws_config: AWSConfig) -> None:
    """
    Validates that given aws credentials are valid. Dependency
    exception is raised if failure occurs.
    """
    import fides.connectors.aws as aws_connector

    aws_connector.validate_credentials(aws_config=aws_config)


async def validate_bigquery(bigquery_config: BigQueryConfig) -> None:
    """
    Validates that given GCP BigQuery credentials are valid. Dependency
    exception is raised if failure occurs.
    """
    import fides.connectors.bigquery as bigquery_connector

    bigquery_engine = bigquery_connector.get_bigquery_engine(bigquery_config)
    bigquery_connector.validate_bigquery_engine(bigquery_engine)


async def validate_okta(okta_config: OktaConfig) -> None:
    """
    Validates that given okta credentials are valid. Dependency
    exception is raised if failure occurs.
    """
    import fides.connectors.okta as okta_connector

    await okta_connector.validate_credentials(okta_config=okta_config)



"""Module that adds models for connectors"""

# pylint: disable=C0115,C0116, E0213
from typing import ClassVar, List, Optional

from pydantic import BaseModel


class AWSConfig(BaseModel):
    """
    The model for the connection config for AWS
    """

    region_name: str
    aws_secret_access_key: str
    aws_access_key_id: str
    aws_session_token: Optional[str] = None


class KeyfileCreds(BaseModel):
    """The model for BigQuery credential keyfiles."""

    type: Optional[str] = None
    project_id: str
    private_key_id: Optional[str] = None
    private_key: Optional[str] = None
    client_email: Optional[str] = None
    client_id: Optional[str] = None
    auth_uri: Optional[str] = None
    token_uri: Optional[str] = None
    auth_provider_x509_cert_url: Optional[str] = None
    client_x509_cert_url: Optional[str] = None


class BigQueryConfig(BaseModel):
    """
    The model for the connection config for BigQuery
    """

    dataset: Optional[str] = None
    keyfile_creds: KeyfileCreds

    _required_components: ClassVar[List[str]] = ["keyfile_creds"]


class OktaConfig(BaseModel):
    """
    The model for the connection config for Okta
    """

    # camel case matches okta client config model
    orgUrl: str
    token: str


class DatabaseConfig(BaseModel):
    """
    The model for the connection config for databases
    """

    connection_string: str


class ConnectorFailureException(Exception):
    """
    Connector exception for unspecified failures
    """


class ConnectorAuthFailureException(Exception):
    """
    Connector exception for authentication failures
    """



from typing import Any, Dict, Optional, Union

import pydash
from loguru import logger
from requests import Response

from fides.api.common_exceptions import FidesopsException
from fides.api.schemas.saas.shared_schemas import ConnectorParamRef, SaaSRequestParams
from fides.api.schemas.saas.strategy_configuration import OffsetPaginationConfiguration
from fides.api.service.pagination.pagination_strategy import PaginationStrategy


class OffsetPaginationStrategy(PaginationStrategy):
    name = "offset"
    configuration_model = OffsetPaginationConfiguration

    def __init__(self, configuration: OffsetPaginationConfiguration):
        self.incremental_param = configuration.incremental_param
        self.increment_by = configuration.increment_by
        self.limit = configuration.limit

    def get_next_request(
        self,
        request_params: SaaSRequestParams,
        connector_params: Dict[str, Any],
        response: Response,
        data_path: Optional[str],
    ) -> Optional[SaaSRequestParams]:
        """Build request for next page of data"""

        # stop paginating if response did not contain any data
        response_data = (
            pydash.get(response.json(), data_path) if data_path else response.json()
        )
        if not response_data:
            return None

        # find query param value from deconstructed request_params, throw exception if query param not found
        param_value = request_params.query_params.get(self.incremental_param)
        if param_value is None:
            raise FidesopsException(
                f"Unable to find query param named '{self.incremental_param}' in request"
            )

        # increment param value and return None if limit has been reached to indicate there are no more pages
        limit: Optional[Union[int, ConnectorParamRef]] = self.limit
        if isinstance(self.limit, ConnectorParamRef):
            limit = connector_params.get(self.limit.connector_param)
            if limit is None:
                raise FidesopsException(
                    f"Unable to find value for 'limit' with the connector_param reference '{self.limit.connector_param}'"
                )
            try:
                limit = int(limit)
            except ValueError:
                raise FidesopsException(
                    f"The value '{limit}' of the '{self.limit.connector_param}' connector_param could not be cast to an int"
                )
        param_value += self.increment_by
        if limit and param_value > limit:
            logger.info("Pagination limit has been reached")
            return None

        # update query param and return updated request_param tuple
        request_params.query_params[self.incremental_param] = param_value
        return SaaSRequestParams(
            method=request_params.method,
            headers=request_params.headers,
            path=request_params.path,
            query_params=request_params.query_params,
            body=request_params.body,
        )

    def validate_request(self, request: Dict[str, Any]) -> None:
        """Ensures that the query param specified by 'incremental_param' exists in the request"""
        query_params = (
            query_params
            for query_params in request.get("query_params", [])
            if query_params.get("name") == self.incremental_param
        )
        query_param = next(query_params, None)
        if query_param is None:
            raise ValueError(f"Query param '{self.incremental_param}' not found.")



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import Any

from citext import CIText
from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session

from fides.api.cryptography.cryptographic_util import (
    generate_salt,
    hash_credential_with_salt,
)
from fides.api.db.base_class import Base

INVITE_CODE_TTL_HOURS = 24


class FidesUserInvite(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "fides_user_invite"

    username = Column(  # type: ignore
        CIText,
        ForeignKey("fidesuser.username", ondelete="CASCADE"),
        primary_key=True,
    )
    hashed_invite_code = Column(String, nullable=False)
    salt = Column(String, nullable=False)

    @classmethod
    def hash_invite_code(
        cls, invite_code: str, encoding: str = "UTF-8"
    ) -> tuple[str, str]:
        """Utility function to hash a user's invite code with a generated salt."""

        salt = generate_salt()
        hashed_invite_code = hash_credential_with_salt(
            invite_code.encode(encoding),
            salt.encode(encoding),
        )
        return hashed_invite_code, salt

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = False
    ) -> FidesUserInvite:
        """
        Create a FidesUserInvite by hashing the invite code with a generated salt
        and storing the hashed invite code and the salt.
        """

        hashed_invite_code, salt = FidesUserInvite.hash_invite_code(data["invite_code"])

        user_invite = super().create(
            db,
            data={
                "username": data["username"],
                "hashed_invite_code": hashed_invite_code,
                "salt": salt,
            },
        )

        return user_invite

    def invite_code_valid(self, invite_code: str, encoding: str = "UTF-8") -> bool:
        """Verifies that the provided invite code is correct."""
        if self.salt is None:
            return False

        invite_code_hash = hash_credential_with_salt(
            invite_code.encode(encoding),
            self.salt.encode(encoding),
        )

        return invite_code_hash == self.hashed_invite_code

    def is_expired(self) -> bool:
        """Check if the invite has expired."""

        current_time_utc = datetime.now(timezone.utc)

        if not self.updated_at:
            return True

        expiration_datetime = self.updated_at + timedelta(hours=INVITE_CODE_TTL_HOURS)
        return current_time_utc > expiration_datetime

    def renew_invite(self, db: Session, new_invite_code: str) -> None:
        """Updates the invite code and extends the expiration."""

        hashed_invite_code, salt = FidesUserInvite.hash_invite_code(new_invite_code)
        self.hashed_invite_code = hashed_invite_code
        self.salt = salt
        self.save(db)



from abc import ABC, abstractmethod
from typing import List, Optional, Type, TypeVar

from fideslang.models import Dataset as FideslangDataset
from sqlalchemy.orm import Session

from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.schemas.dataset import DatasetTraversalDetails, ValidateDatasetResponse

T = TypeVar("T", bound="DatasetValidationStep")


class DatasetValidationContext:
    """Context object holding state for validation"""

    def __init__(
        self,
        db: Session,
        dataset: FideslangDataset,
        connection_config: Optional[ConnectionConfig] = None,
    ):
        self.db = db
        self.dataset = dataset
        self.connection_config = connection_config
        self.traversal_details: Optional[DatasetTraversalDetails] = None


class DatasetValidationStep(ABC):
    """Abstract base class for dataset validation steps"""

    @classmethod
    def _find_all_validation_steps(cls: Type[T]) -> List[Type[T]]:
        """Find all subclasses of DatasetValidationStep"""
        return cls.__subclasses__()

    @abstractmethod
    def validate(self, context: DatasetValidationContext) -> None:
        """Perform validation step"""


class DatasetValidator:
    """Dataset validator that allows selective validation steps"""

    def __init__(
        self,
        db: Session,
        dataset: FideslangDataset,
        connection_config: Optional[ConnectionConfig] = None,
        skip_steps: Optional[List[Type[DatasetValidationStep]]] = None,
    ):
        self.context = DatasetValidationContext(db, dataset, connection_config)
        self.skip_steps = skip_steps or []
        self.validation_steps = [
            step()
            for step in DatasetValidationStep._find_all_validation_steps()
            if step not in self.skip_steps
        ]

    def validate(self) -> ValidateDatasetResponse:
        """Execute selected validation steps and return response"""
        for step in self.validation_steps:
            step.validate(self.context)

        return ValidateDatasetResponse(
            dataset=self.context.dataset,
            traversal_details=self.context.traversal_details,
        )



"""This module defines the base settings for the other settings to inherit from."""

# pylint: disable=C0115,C0116, E0213

from typing import Optional, Tuple, Type

from pydantic import ValidationInfo
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)


class FidesSettings(BaseSettings):
    """Class used as a base model for configuration subsections."""

    model_config = SettingsConfigDict(extra="allow")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """Set environment variables to take precedence over init values."""
        return env_settings, init_settings, file_secret_settings


def port_integer_converter(
    info: ValidationInfo, port_name: Optional[str] = None
) -> int:
    """
    Convert supplied port value to an integer.

    "Before" field validators mean that port is not yet guaranteed to exist or be the appropriate type
    """
    port = info.data.get(port_name or "port")

    if not isinstance(port, (str, int)):  # This check also helps satisfy mypy
        raise ValueError(
            "Port must be supplied and able to be converted to an integer."
        )

    return int(port)



"""add custom report table

Revision ID: c9abc9bbaa49
Revises: fa476b733915
Create Date: 2024-09-26 23:19:58.766385

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "c9abc9bbaa49"
down_revision = "fa476b733915"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "plus_custom_report",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("type", sa.String(), nullable=False),
        sa.Column("created_by", sa.String(), nullable=True),
        sa.Column("config", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.ForeignKeyConstraint(["created_by"], ["fidesuser.id"], ondelete="SET NULL"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_index(
        op.f("ix_plus_custom_report_id"), "plus_custom_report", ["id"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_plus_custom_report_id"), table_name="plus_custom_report")
    op.drop_table("plus_custom_report")



import hashlib
import os.path
import tempfile
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Dict, Iterator, Optional, Type
from urllib.request import urlretrieve

from botocore.client import BaseClient
from loguru import logger
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine

from fides.api.schemas.connection_configuration.connection_secrets_base_rds import (
    BaseRDSSchema,
)
from fides.api.util.aws_util import get_aws_session

CA_CERT_URL = "https://truststore.pki.rds.amazonaws.com/global/global-bundle.pem"


class RDSConnectorMixin(ABC):

    @property
    @abstractmethod
    def url_scheme(self) -> str:
        """
        Returns the URL scheme for the connector's Engine.
        """

    @property
    @abstractmethod
    def typed_secrets(self) -> BaseRDSSchema:
        """
        Returns a strongly typed secrets object.
        """

    @property
    @abstractmethod
    def aws_engines(self) -> list[str]:
        """
        Returns the AWS engines supported by the connector.
        """

    @cached_property
    def global_bundle_uri(self) -> str:
        """
        Returns the global bundle for the connector.
        """
        logger.info("Getting RDS CA cert bundle")
        tempdir = tempfile.gettempdir()
        url_hash = hashlib.sha256(CA_CERT_URL.encode()).hexdigest()
        local_file_name = f"fides_rds_ca_cert_{url_hash}.pem"
        bundle_uri = os.path.join(tempdir, local_file_name)
        if not os.path.isfile(bundle_uri):
            logger.info("Downloading RDS CA cert bundle")
            urlretrieve(CA_CERT_URL, bundle_uri)
        logger.info(f"Using RDS CA cert bundle: {bundle_uri}")
        return bundle_uri

    def get_connect_args(self) -> Dict:
        """
        Returns the connection arguments for the Engine.
        """
        return {}

    @cached_property
    def rds_client(self) -> Type[BaseClient]:
        """
        Returns an RDS client.
        """
        logger.info("Creating RDS client")
        secrets = self.typed_secrets.model_dump()
        secrets["region_name"] = self.typed_secrets.region
        return get_aws_session(
            self.typed_secrets.auth_method,
            secrets,  # type: ignore[arg-type]
            self.typed_secrets.aws_assume_role_arn,
        ).client("rds")

    def get_authentication_token(self, host: str, port: str, user: str) -> str:
        """
        Returns the authentication token for the provided host, port, and user.
        """
        logger.info(
            f"Generating DB auth token for ({host}:{port}) user ({user}) and region ({self.typed_secrets.region})"
        )
        return self.rds_client.generate_db_auth_token(
            host,
            port,
            user,
            self.typed_secrets.region,
        )

    def get_database_instances_connection_info(self) -> Iterator[dict]:
        """
        Returns the connection info for all database instances.
        """
        rds_client = self.rds_client

        # Get all RDS instances
        logger.info(f"Listing RDS {self.aws_engines} instances")
        instances = rds_client.describe_db_instances(
            Filters=[
                {"Name": "engine", "Values": self.aws_engines},
            ]
        )

        for instance in instances["DBInstances"]:
            if "DBClusterIdentifier" in instance:
                # There is a case where Endpoint is not available for DB Clusters
                logger.info(
                    f"Found a DB Cluster, skipping: {instance['DBClusterIdentifier']}"
                )
                continue

            if not instance["IAMDatabaseAuthenticationEnabled"]:
                # There is a case where IAM Authentication is not enabled
                logger.info(
                    f"Skipping DB Instance, because IAMDatabaseAuthenticationEnabled=False: {instance['DBInstanceIdentifier']}"
                )
                continue

            logger.info(
                f"Found RDS {self.aws_engines} instance: {instance['DBInstanceIdentifier']}"
            )
            yield {
                "name": instance["DBInstanceIdentifier"],  # Str
                "host": instance["Endpoint"]["Address"],  # Str
                "port": instance["Endpoint"]["Port"],  # Int
            }

        # Get all RDS clusters
        logger.info(f"Listing RDS {self.aws_engines} clusters")
        instances = rds_client.describe_db_clusters(
            Filters=[
                {"Name": "engine", "Values": self.aws_engines},
            ]
        )

        for instance in instances["DBClusters"]:
            if not instance["IAMDatabaseAuthenticationEnabled"]:
                # There is a case where IAM Authentication is not enabled
                logger.info(
                    f"Skipping DB Cluster, because IAMDatabaseAuthenticationEnabled=False: {instance['DBClusterIdentifier']}"
                )
                continue

            logger.info(
                f"Found RDS {self.aws_engines} cluster: {instance['DBClusterIdentifier']}"
            )
            yield {
                "name": instance["DBClusterIdentifier"],
                "host": instance["Endpoint"],
                "port": instance["Port"],
            }

    @cached_property
    def database_instances_connection_info(self) -> dict[str, dict]:
        """
        Returns the cached connection info for all database instances.
        {
            "db_instance_name": {
                "name": "db_instance_name",
                "host": "host",
                "port": "port",
            }
        }
        """
        instances_info = {
            info["name"]: info for info in self.get_database_instances_connection_info()
        }
        logger.info(f"Getting instances connection info: {instances_info}")
        return instances_info

    def get_database_instance_connection_info(
        self, database_instance_name: str
    ) -> dict[str, str]:
        """
        Retrieves the connection info for the provided database instance name.
        """
        database_instance_connection_info = self.database_instances_connection_info.get(
            database_instance_name
        )

        if not database_instance_connection_info:
            raise ValueError(f"Database instance '{database_instance_name}' not found!")

        return database_instance_connection_info

    def create_engine(
        self, db_username: str, host: str, port: str, db_name: Optional[str] = None
    ) -> Engine:
        """
        Returns a SQLAlchemy Engine that can be used to interact with a database
        """

        url = f"{self.url_scheme}://{db_username}@{host}:{port}" + (
            f"/{db_name}" if db_name else ""
        )

        connect_args = self.get_connect_args()

        logger.info(f"Creating SQLAlchemy engine for {url}")

        engine = create_engine(url, connect_args=connect_args)

        @event.listens_for(engine, "do_connect")
        def provide_token(dialect, conn_rec, cargs, cparams):  # type: ignore[no-untyped-def]
            """
            Provide the authentication token which has to be created under demand. It expires after 15 minutes.
            """
            cparams["password"] = self.get_authentication_token(host, port, db_username)

        return engine



"""Add fidesctl_meta to System

Revision ID: edcd28ede1f7
Revises: 50b3736634d2
Create Date: 2022-02-17 00:41:41.595166

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "edcd28ede1f7"
down_revision = "50b3736634d2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("systems", sa.Column("fidesctl_meta", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("systems", "fidesctl_meta")
    # ### end Alembic commands ###



"""Updating Attentive connection to Attentive email

Revision ID: 59d9c4beeb30
Revises: 75bb9ee843f5
Create Date: 2024-09-24 14:52:49.199323

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "59d9c4beeb30"
down_revision = "75bb9ee843f5"
branch_labels = None
depends_on = None


def upgrade():
    # Note: Since we may have live references to attentive
    # add 'attentive_email' to ConnectionType enum
    op.execute("ALTER TYPE connectiontype RENAME TO connectiontype_old")
    op.execute(
        """
        CREATE TYPE connectiontype AS ENUM (
            'mongodb',
            'mysql',
            'https',
            'snowflake',
            'redshift',
            'mssql',
            'mariadb',
            'bigquery',
            'saas',
            'manual',
            'manual_webhook',
            'timescale',
            'fides',
            'sovrn',
            'attentive',
            'attentive_email',
            'dynamodb',
            'postgres',
            'generic_consent_email',
            'generic_erasure_email',
            'scylla',
            's3',
            'google_cloud_sql_mysql',
            'google_cloud_sql_postgres',
            'dynamic_erasure_email'
            'rds_mysql'
        )
    """
    )
    op.execute(
        """
        ALTER TABLE connectionconfig ALTER COLUMN connection_type TYPE connectiontype USING
        connection_type::text::connectiontype
    """
    )
    op.execute("DROP TYPE connectiontype_old")

    # update connectionconfig reference from 'attentive' to 'attentive_email'
    op.execute(
        """
        UPDATE connectionconfig
        SET connection_type = 'attentive_email'::connectiontype
        WHERE connection_type = 'attentive'::connectiontype
        """
    )

    # remove 'attentive' from ConnectionType enum
    op.execute("ALTER TYPE connectiontype RENAME TO connectiontype_staging")

    op.execute(
        """
        CREATE TYPE connectiontype AS ENUM (
            'mongodb',
            'mysql',
            'https',
            'snowflake',
            'redshift',
            'mssql',
            'mariadb',
            'bigquery',
            'saas',
            'manual',
            'manual_webhook',
            'timescale',
            'fides',
            'sovrn',
            'attentive_email',
            'dynamodb',
            'postgres',
            'generic_consent_email',
            'generic_erasure_email',
            'scylla',
            's3',
            'google_cloud_sql_mysql',
            'google_cloud_sql_postgres',
            'dynamic_erasure_email',
            'rds_mysql'
        )
    """
    )
    op.execute(
        """
        ALTER TABLE connectionconfig ALTER COLUMN connection_type TYPE connectiontype USING
        connection_type::text::connectiontype
    """
    )
    op.execute("DROP TYPE connectiontype_staging")


def downgrade():
    # add 'attentive' to ConnectionType enum
    op.execute("ALTER TYPE connectiontype RENAME TO connectiontype_old")
    op.execute(
        """
        CREATE TYPE connectiontype AS ENUM (
            'mongodb',
            'mysql',
            'https',
            'snowflake',
            'redshift',
            'mssql',
            'mariadb',
            'bigquery',
            'saas',
            'manual',
            'manual_webhook',
            'timescale',
            'fides',
            'sovrn',
            'attentive',
            'attentive_email',
            'dynamodb',
            'postgres',
            'generic_consent_email',
            'generic_erasure_email',
            'scylla',
            's3',
            'google_cloud_sql_mysql',
            'google_cloud_sql_postgres',
            'dynamic_erasure_email',
            'rds_mysql'
        )
    """
    )
    op.execute(
        """
        ALTER TABLE connectionconfig ALTER COLUMN connection_type TYPE connectiontype USING
        connection_type::text::connectiontype
    """
    )
    op.execute("DROP TYPE connectiontype_old")

    # update connectionconfig reference from 'attentive_email' to 'attentive'
    op.execute(
        """
        UPDATE connectionconfig
        SET connection_type = 'attentive'::connectiontype
        WHERE connection_type = 'attentive_email'::connectiontype
        """
    )

    # remove 'attentive_email' from ConnectionType enum
    op.execute("ALTER TYPE connectiontype RENAME TO connectiontype_staging")

    op.execute(
        """
        CREATE TYPE connectiontype AS ENUM (
            'mongodb',
            'mysql',
            'https',
            'snowflake',
            'redshift',
            'mssql',
            'mariadb',
            'bigquery',
            'saas',
            'manual',
            'manual_webhook',
            'timescale',
            'fides',
            'sovrn',
            'attentive',
            'dynamodb',
            'postgres',
            'generic_consent_email',
            'generic_erasure_email',
            'scylla',
            's3',
            'google_cloud_sql_mysql',
            'google_cloud_sql_postgres',
            'dynamic_erasure_email'
        )
    """
    )

    op.execute(
        """
        ALTER TABLE connectionconfig ALTER COLUMN connection_type TYPE connectiontype USING
        connection_type::text::connectiontype
    """
    )
    op.execute("DROP TYPE connectiontype_staging")



from typing import Dict, Optional

from google.cloud.storage import Client  # type: ignore
from google.oauth2 import service_account
from loguru import logger

from fides.api.common_exceptions import StorageUploadError
from fides.api.schemas.storage.storage import GCSAuthMethod


def get_gcs_client(
    auth_method: str,
    storage_secrets: Optional[Dict],
) -> Client:
    """
    Abstraction to retrieve a GCS client using secrets.
    """
    if auth_method == GCSAuthMethod.ADC.value:
        storage_client = Client()

    elif auth_method == GCSAuthMethod.SERVICE_ACCOUNT_KEYS.value:
        if not storage_secrets:
            err_msg = "Storage secrets not found for Google Cloud Storage."
            logger.warning(err_msg)
            raise StorageUploadError(err_msg)

        credentials = service_account.Credentials.from_service_account_info(
            dict(storage_secrets)
        )
        storage_client = Client(credentials=credentials)

    else:
        logger.error("Google Cloud Storage auth method not supported: {}", auth_method)
        raise ValueError(
            f"Google Cloud Storage auth method not supported: {auth_method}"
        )

    return storage_client



"""purpose_header_to_translations

Adds new purpose_header field to Experience Translation table and
Privacy Experience Config History table

Revision ID: 1332815dcd71
Revises: 9cad5a5c438c
Create Date: 2024-08-28 21:05:41.933572

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1332815dcd71"
down_revision = "9cad5a5c438c"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "experiencetranslation", sa.Column("purpose_header", sa.String(), nullable=True)
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("purpose_header", sa.String(), nullable=True),
    )


def downgrade():
    op.drop_column("privacyexperienceconfighistory", "purpose_header")
    op.drop_column("experiencetranslation", "purpose_header")



"""track experience on preferences

Revision ID: 2661f31daffb
Revises: b546ce845e6c
Create Date: 2023-05-16 20:05:29.478005

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "2661f31daffb"
down_revision = "b546ce845e6c"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "anonymized_ip_address",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
    )
    op.add_column(
        "privacypreferencehistory", sa.Column("method", sa.String(), nullable=True)
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("privacy_experience_config_history_id", sa.String(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("privacy_experience_history_id", sa.String(), nullable=True),
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_privacy_experience_config_history_id"),
        "privacypreferencehistory",
        ["privacy_experience_config_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_privacy_experience_history_id"),
        "privacypreferencehistory",
        ["privacy_experience_history_id"],
        unique=False,
    )
    op.create_foreign_key(
        "privacypreferencehistory_privacy_experience_history_id_fkey",
        "privacypreferencehistory",
        "privacyexperiencehistory",
        ["privacy_experience_history_id"],
        ["id"],
    )
    op.create_foreign_key(
        "privacypreferencehistory_privacy_experience_config_history_fkey",
        "privacypreferencehistory",
        "privacyexperienceconfighistory",
        ["privacy_experience_config_history_id"],
        ["id"],
    )


def downgrade():
    op.drop_constraint(
        "privacypreferencehistory_privacy_experience_history_id_fkey",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.drop_constraint(
        "privacypreferencehistory_privacy_experience_config_history_fkey",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_privacy_experience_history_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_privacy_experience_config_history_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_column("privacypreferencehistory", "privacy_experience_history_id")
    op.drop_column("privacypreferencehistory", "privacy_experience_config_history_id")
    op.drop_column("privacypreferencehistory", "method")
    op.drop_column("privacypreferencehistory", "anonymized_ip_address")



import json
from typing import ClassVar, List, Optional, Union

from google.cloud.bigquery import Client as BigQueryClient
from pydantic import EmailStr, Field, field_validator
from pydantic.main import BaseModel

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


class KeyfileCreds(BaseModel):
    """Schema that holds BigQuery keyfile key/vals"""

    type: Optional[str] = None
    project_id: str = Field(title="Project ID")
    private_key_id: Optional[str] = Field(default=None, title="Private key ID")
    private_key: Optional[str] = Field(
        default=None, title="Private key", json_schema_extra={"sensitive": True}
    )
    client_email: Optional[EmailStr] = Field(None, title="Client email")
    client_id: Optional[str] = Field(default=None, title="Client ID")
    auth_uri: Optional[str] = Field(default=None, title="Auth URI")
    token_uri: Optional[str] = Field(default=None, title="Token URI")
    auth_provider_x509_cert_url: Optional[str] = Field(
        default=None, title="Auth provider X509 cert URL"
    )
    client_x509_cert_url: Optional[str] = Field(
        default=None, title="Client X509 cert URL"
    )


class BigQuerySchema(ConnectionConfigSecretsSchema):
    """Schema to validate the secrets needed to connect to BigQuery"""

    keyfile_creds: KeyfileCreds = Field(
        title="Keyfile creds",
        json_schema_extra={"sensitive": True},
        description="The contents of the key file that contains authentication credentials for a service account in GCP.",
    )
    dataset: Optional[str] = Field(
        default=None,
        title="Dataset",
        description="Only provide a dataset to scope discovery monitors and privacy request automation to a specific BigQuery dataset. In most cases, this can be left blank.",
    )

    _required_components: ClassVar[List[str]] = ["keyfile_creds"]

    @field_validator("keyfile_creds", mode="before")
    @classmethod
    def parse_keyfile_creds(cls, v: Union[str, dict]) -> KeyfileCreds:
        if isinstance(v, str):
            v = json.loads(v)
        return KeyfileCreds.model_validate(v)

    def get_client(self) -> BigQueryClient:
        return BigQueryClient.from_service_account_info(
            self.keyfile_creds.model_dump()  # pylint: disable=no-member
        )


class BigQueryDocsSchema(BigQuerySchema, NoValidationSchema):
    """BigQuery Secrets Schema for API Docs"""



"""adds MAILCHIMP_TRANSACTIONAL enum

Revision ID: 39b209861471
Revises: 9f38dad37628
Create Date: 2023-03-03 21:57:14.368385

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "39b209861471"
down_revision = "9f38dad37628"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type messagingservicetype rename to messagingservicetype_old")
    op.execute(
        "create type messagingservicetype as enum('MAILCHIMP_TRANSACTIONAL', 'MAILGUN', 'TWILIO_TEXT', 'TWILIO_EMAIL')"
    )
    op.execute(
        (
            "alter table messagingconfig alter column service_type type messagingservicetype using "
            "service_type::text::messagingservicetype"
        )
    )
    op.execute("drop type messagingservicetype_old")


def downgrade():
    # return
    op.execute(
        "delete from messagingconfig where service_type in ('MAILCHIMP_TRANSACTIONAL')"
    )
    op.execute("alter type messagingservicetype rename to messagingservicetype_old")
    op.execute(
        "create type messagingservicetype as enum('MAILGUN', 'TWILIO_TEXT', 'TWILIO_EMAIL')"
    )
    op.execute(
        (
            "alter table messagingconfig alter column service_type type messagingservicetype using "
            "service_type::text::messagingservicetype"
        )
    )
    op.execute("drop type messagingservicetype_old")



from sqlalchemy import Boolean, CheckConstraint, Column, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import Base  # type: ignore[attr-defined]
from fides.api.db.base_class import JSONTypeOverride
from fides.config import CONFIG


class IdentitySalt(Base):
    """
    A single-row table used to store the encrypted salt value used for identity hashing (SHA-256).
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "identity_salt"

    encrypted_value = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=True,
    )  # Type bytea in the db
    single_row = Column(
        Boolean,
        default=True,
        nullable=False,
        unique=True,
    )  # used to constrain table to one row

    CheckConstraint("single_row", name="identity_salt_single_row_check")
    UniqueConstraint("single_row", name="identity_salt_single_row_unique")



"""reduce experience and config duplication

Revision ID: 317e6197c76a
Revises: b8248ce6b8a1
Create Date: 2023-06-05 23:24:43.174826

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "317e6197c76a"
down_revision = "b8248ce6b8a1"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index(
        "ix_privacyexperiencehistory_experience_config_history_id",
        table_name="privacyexperiencehistory",
    )
    op.drop_index(
        "ix_privacyexperiencehistory_experience_config_id",
        table_name="privacyexperiencehistory",
    )
    op.drop_index(
        "ix_privacyexperiencehistory_id", table_name="privacyexperiencehistory"
    )
    op.drop_index(
        "ix_privacyexperiencehistory_privacy_experience_id",
        table_name="privacyexperiencehistory",
    )
    op.drop_index(
        "ix_privacyexperiencehistory_region", table_name="privacyexperiencehistory"
    )
    op.drop_constraint(
        "privacypreferencehistory_privacy_experience_history_id_fkey",
        "privacypreferencehistory",
        type_="foreignkey",
    )

    op.drop_table("privacyexperiencehistory")

    op.drop_index(
        "ix_privacyexperience_experience_config_history_id",
        table_name="privacyexperience",
    )
    op.drop_constraint(
        "privacyexperience_experience_config_history_id_fkey",
        "privacyexperience",
        type_="foreignkey",
    )

    op.drop_column("privacyexperience", "disabled")
    op.drop_column("privacyexperience", "version")
    op.drop_column("privacyexperience", "experience_config_history_id")

    op.drop_index(
        "ix_privacypreferencehistory_privacy_experience_history_id",
        table_name="privacypreferencehistory",
    )
    op.drop_column("privacypreferencehistory", "privacy_experience_history_id")

    op.add_column(
        "privacypreferencehistory",
        sa.Column("privacy_experience_id", sa.String(), nullable=True),
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_privacy_experience_id"),
        "privacypreferencehistory",
        ["privacy_experience_id"],
        unique=False,
    )
    op.create_foreign_key(
        "privacypreferencehistory_privacy_experience_id_fkey",
        "privacypreferencehistory",
        "privacyexperience",
        ["privacy_experience_id"],
        ["id"],
    )
    # ### end Alembic commands ###


def downgrade():
    op.create_table(
        "privacyexperiencehistory",
        sa.Column("id", sa.VARCHAR(length=255), autoincrement=False, nullable=False),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("disabled", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column("component", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column(
            "version",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            "privacy_experience_id", sa.VARCHAR(), autoincrement=False, nullable=False
        ),
        sa.Column("region", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column(
            "experience_config_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "experience_config_history_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["experience_config_history_id"],
            ["privacyexperienceconfighistory.id"],
            name="privacyexperiencehistory_experience_config_history_id_fkey",
        ),
        sa.ForeignKeyConstraint(
            ["experience_config_id"],
            ["privacyexperienceconfig.id"],
            name="privacyexperiencehistory_experience_config_id_fkey",
        ),
        sa.ForeignKeyConstraint(
            ["privacy_experience_id"],
            ["privacyexperience.id"],
            name="privacyexperiencehistory_privacy_experience_id_fkey",
        ),
        sa.PrimaryKeyConstraint("id", name="privacyexperiencehistory_pkey"),
    )

    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "privacy_experience_history_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )

    op.create_foreign_key(
        "privacypreferencehistory_privacy_experience_history_id_fkey",
        "privacypreferencehistory",
        "privacyexperiencehistory",
        ["privacy_experience_history_id"],
        ["id"],
    )

    op.drop_index(
        op.f("ix_privacypreferencehistory_privacy_experience_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_constraint(
        "privacypreferencehistory_privacy_experience_id_fkey",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.drop_column("privacypreferencehistory", "privacy_experience_id")

    op.create_index(
        "ix_privacypreferencehistory_privacy_experience_history_id",
        "privacypreferencehistory",
        ["privacy_experience_history_id"],
        unique=False,
    )

    op.add_column(
        "privacyexperience",
        sa.Column(
            "experience_config_history_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "version",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "disabled",
            sa.BOOLEAN(),
            server_default="t",
            autoincrement=False,
            nullable=False,
        ),
    )
    op.create_foreign_key(
        "privacyexperience_experience_config_history_id_fkey",
        "privacyexperience",
        "privacyexperienceconfighistory",
        ["experience_config_history_id"],
        ["id"],
    )

    op.create_index(
        "ix_privacyexperience_experience_config_history_id",
        "privacyexperience",
        ["experience_config_history_id"],
        unique=False,
    )
    op.create_index(
        "ix_privacyexperiencehistory_region",
        "privacyexperiencehistory",
        ["region"],
        unique=False,
    )
    op.create_index(
        "ix_privacyexperiencehistory_privacy_experience_id",
        "privacyexperiencehistory",
        ["privacy_experience_id"],
        unique=False,
    )
    op.create_index(
        "ix_privacyexperiencehistory_id",
        "privacyexperiencehistory",
        ["id"],
        unique=False,
    )
    op.create_index(
        "ix_privacyexperiencehistory_experience_config_id",
        "privacyexperiencehistory",
        ["experience_config_id"],
        unique=False,
    )
    op.create_index(
        "ix_privacyexperiencehistory_experience_config_history_id",
        "privacyexperiencehistory",
        ["experience_config_history_id"],
        unique=False,
    )



"""add_system_vendor_id

Revision ID: 2af81f2b1a6f
Revises: b3b407ce2d5f
Create Date: 2023-11-06 21:46:30.715273

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2af81f2b1a6f"
down_revision = "b3b407ce2d5f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "ctl_systems", sa.Column("previous_vendor_id", sa.String(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("ctl_systems", "previous_vendor_id")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import Callable, Generic, Optional, TypeVar

T = TypeVar("T")


class MatchingQueue(Generic[T]):
    """A basic LILO queue with the added ability to pop not only the head, but the first value matching a given input function."""

    def __init__(self, *values: T):
        self.data = list(values)

    def push(self, t: T) -> None:
        """insert into the queue"""
        self.data.append(t)

    def push_if_new(self, t: T) -> None:
        """insert into the queue only if this value is not already in the queue."""
        if t not in self.data:
            self.push(t)

    def pop(self) -> Optional[T]:
        """safely return the next value in the queue, or None if the queue is empty."""
        if self.data:
            v = self.data[0]
            del self.data[0]
            return v
        return None

    def pop_first_match(self, fn: Callable[[T], bool]) -> Optional[T]:
        """Find the first matching value for f and pop it or return None"""
        for idx, val in enumerate(self.data):
            if fn(val):
                del self.data[idx]
                return val
        # if no matching value exists, return None
        return None

    def is_empty(self) -> bool:
        """is the queue empty?"""
        return len(self.data) == 0

    def __repr__(self) -> str:
        return f"Queue {self.data}"



"""add config_set column to application settings

Revision ID: c9ee230fa6da
Revises: 8e198eb13802
Create Date: 2023-02-01 15:13:52.133075

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy import text
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.sql.elements import TextClause
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import JSONTypeOverride
from fides.config import CONFIG

# revision identifiers, used by Alembic.
revision = "c9ee230fa6da"
down_revision = "8e198eb13802"
branch_labels = None
depends_on = None


def upgrade():
    encryptor = sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(
        JSONTypeOverride,
        CONFIG.security.app_encryption_key,
        AesGcmEngine,
        "pkcs5",
    )
    empty_obj = encryptor.process_bind_param({}, JSON)

    op.add_column(
        "applicationconfig",
        sa.Column(
            "config_set",
            StringEncryptedType(),
            nullable=False,
            server_default=empty_obj,
        ),
    )

    # reset the server_default now that it's initial value has been set
    op.alter_column(
        "applicationconfig", "config_set", nullable=False, server_default=None
    )

    # include this update here to make up for an earlier miss when creating the table
    op.alter_column("applicationconfig", "api_set", nullable=False)


def downgrade():
    op.drop_column("applicationconfig", "config_set")

    # add a downgrade here for consistency
    op.alter_column("applicationconfig", "api_set", nullable=True)



from __future__ import annotations

from typing import Any, Dict, Set

from fides.api.common_exceptions import NoSuchConnectionTypeSecretSchemaError
from fides.api.models.connectionconfig import ConnectionType
from fides.api.schemas.connection_configuration import (
    SaaSSchemaFactory,
    secrets_schemas,
)
from fides.api.schemas.connection_configuration.connection_type_system_map import (
    ConnectionSystemTypeMap,
)
from fides.api.schemas.connection_configuration.enums.system_type import SystemType
from fides.api.schemas.policy import SUPPORTED_ACTION_TYPES, ActionType
from fides.api.schemas.saas.saas_config import SaaSConfig
from fides.api.service.connectors.consent_email_connector import (
    CONSENT_EMAIL_CONNECTOR_TYPES,
)
from fides.api.service.connectors.erasure_email_connector import (
    ERASURE_EMAIL_CONNECTOR_TYPES,
)
from fides.api.service.connectors.saas.connector_registry_service import (
    ConnectorRegistry,
)
from fides.api.util.saas_util import load_config_from_string


def transform_v2_to_v1_in_place(schema: Dict[str, Any]) -> None:
    """Transform connection secrets from Pydantic V2 format to V1 format for backwards compatibility
    since Connection secrets UI is built off of this data.

    This is error prone and is subject to change as we add more types of schemas.
    We should consider a refactor of both the frontend and backend here.
    """

    def swap_defs_with_definitions(defn: str) -> str:
        """Reverting to v1 schema format for definitions"""
        return defn.replace("#/$defs", "#/definitions")

    def transform_any_of(field_attributes_mapping: Dict[str, Any]) -> None:
        for attributes in field_attributes_mapping.values():
            # Transforming Pydantic V2 schemas -> Pydantic V1 schemas
            if attributes.get("anyOf"):
                anyOf = attributes.get("anyOf")
                for type_annotation in anyOf:
                    # If field has a null default, V2 returns multiple types in a list
                    # of anyOf which isn't compatible with current UI.  Returning
                    # the first non-null type instead.
                    #
                    # BEFORE
                    # "test_email_address": {
                    #  "anyOf": [
                    #          {
                    #           "format": "email",
                    #           "type": "string"
                    #         },
                    #         {
                    #           "type": "null"
                    #         }
                    #       ]
                    #   }
                    # AFTER
                    #   "test_email_address": {
                    #       "type": "string",
                    #       "format": "email"
                    #   },
                    if "type" in type_annotation and type_annotation["type"] != "null":
                        for key, val in type_annotation.items():
                            attributes[key] = val
                        break
                    # Nested schemas have more complex formatting, they reference a
                    # more detailed definition elsewhere in the schema
                    # Advanced settings don't appear to be connected in the UI.
                    # BEFORE
                    #   "advanced_settings": {
                    #       "anyOf": [
                    #             {
                    #               "$ref": "#/$defs/AdvancedSettings"
                    #             },
                    #             {
                    #               "type": "null"
                    #             }
                    #       ]
                    #   }
                    # AFTER
                    #     "advanced_settings": {
                    #       "title": "Advanced Settings",
                    #       "default": {
                    #         "identity_types": {
                    #           "email": true,
                    #           "phone_number": false
                    #         }
                    #       },
                    #       "allOf": [
                    #         {
                    #           "$ref": "#/definitions/AdvancedSettings"
                    #         }
                    #       ]
                    #     }
                    if "$ref" in type_annotation:
                        attributes["allOf"] = [
                            {
                                "$ref": swap_defs_with_definitions(
                                    type_annotation["$ref"]
                                )
                            }
                        ]
                        break

                attributes.pop("anyOf")

            if "default" in attributes and attributes["default"] is None:
                # Backwards compatible with UI, this affects what fields
                # show up as required
                attributes.pop("default")

            if attributes.get("$ref"):
                # V1 called it "#/$defs", V2 dalls it "#/definitions/"
                attributes["$ref"] = swap_defs_with_definitions(attributes["$ref"])

    transform_any_of(schema["properties"])

    for attributes in schema["properties"].values():
        if attributes.get("allOf"):
            for defs in attributes.get("allOf"):
                for key, val in defs.items():
                    defs[key] = swap_defs_with_definitions(val)

    if schema.get("$defs"):
        schema["definitions"] = schema.get("$defs")
        schema.pop("$defs")
        for key, definition in schema["definitions"].items():
            if definition.get("properties"):
                transform_any_of(definition["properties"])


def get_connection_type_secret_schema(*, connection_type: str) -> dict[str, Any]:
    """Returns the secret fields that should be supplied to authenticate with a particular connection type.

    Note that this does not return actual secrets, instead we return the *types* of
    secret fields needed to authenticate.
    """
    connection_system_types: list[ConnectionSystemTypeMap] = get_connection_types()
    if not any(item.identifier == connection_type for item in connection_system_types):
        raise NoSuchConnectionTypeSecretSchemaError(
            f"No connection type found with name '{connection_type}'."
        )

    if connection_type in [db_type.value for db_type in ConnectionType]:
        schema = secrets_schemas[connection_type].schema()
        transform_v2_to_v1_in_place(schema)
        return schema

    connector_template = ConnectorRegistry.get_connector_template(connection_type)
    if not connector_template:
        raise NoSuchConnectionTypeSecretSchemaError(
            f"No SaaS connector type found with name '{connection_type}'."
        )

    config = SaaSConfig(**load_config_from_string(connector_template.config))
    schema = SaaSSchemaFactory(config).get_saas_schema().schema()
    transform_v2_to_v1_in_place(schema)

    # rearrange the default order of the properties generated by Pydantic
    # to reflect the order defined in the connector_params and external_references
    order = [connector_param.name for connector_param in config.connector_params]
    if config.external_references:
        order.extend(
            [
                external_reference.name
                for external_reference in config.external_references  # pylint:disable=not-an-iterable
            ]
        )

    schema["properties"] = {
        name: value
        for name, value in sorted(
            schema["properties"].items(), key=lambda item: order.index(item[0])
        )
    }
    if schema.get("additionalProperties"):
        schema.pop("additionalProperties")

    # set an empty array for the 'required' key, if there are no required fields on the schema
    # this helps the FE, which is expecting _some_ value here, even if it's an empty list
    if "required" not in schema:
        schema["required"] = []
    return schema


def get_connection_types(
    search: str | None = None,
    system_type: SystemType | None = None,
    action_types: Set[ActionType] = SUPPORTED_ACTION_TYPES,
) -> list[ConnectionSystemTypeMap]:
    def is_match(elem: str) -> bool:
        """If a search query param was included, is it a substring of an available connector type?"""
        return search.lower() in elem.lower() if search else True

    def saas_request_type_filter(connection_type: str) -> bool:
        """
        If any of the request type filters are set to true,
        ensure the given saas connector supports requests of at least one of those types.
        """
        if SUPPORTED_ACTION_TYPES == action_types:
            # if none of our filters are enabled, pass quickly to avoid unnecessary overhead
            return True

        template = ConnectorRegistry.get_connector_template(connection_type)
        if template is None:  # shouldn't happen, but we can be safe
            return False

        # Check if the necessary actions are supported
        return any(
            action_type in template.supported_actions for action_type in action_types
        )

    connection_system_types: list[ConnectionSystemTypeMap] = []
    if (system_type == SystemType.database or system_type is None) and (
        ActionType.access in action_types or ActionType.erasure in action_types
    ):
        database_types: list[ConnectionType] = sorted(
            [
                conn_type
                for conn_type in ConnectionType
                if conn_type
                not in [
                    ConnectionType.attentive_email,
                    ConnectionType.fides,
                    ConnectionType.generic_consent_email,
                    ConnectionType.generic_erasure_email,
                    ConnectionType.dynamic_erasure_email,
                    ConnectionType.https,
                    ConnectionType.manual,
                    ConnectionType.manual_webhook,
                    ConnectionType.saas,
                    ConnectionType.sovrn,
                ]
                and is_match(conn_type.value)
            ],
            key=lambda x: x.value,
        )
        connection_system_types.extend(
            [
                ConnectionSystemTypeMap(
                    identifier=item,
                    type=item.system_type,
                    human_readable=item.human_readable,
                    supported_actions=[ActionType.access, ActionType.erasure],
                )
                for item in database_types
            ]
        )
    if system_type == SystemType.saas or system_type is None:
        saas_types: list[str] = sorted(
            [
                saas_type
                for saas_type in ConnectorRegistry.connector_types()
                if is_match(saas_type) and saas_request_type_filter(saas_type)
            ]
        )

        for item in saas_types:
            connector_template = ConnectorRegistry.get_connector_template(item)
            if connector_template is not None:
                connection_system_types.append(
                    ConnectionSystemTypeMap(
                        identifier=item,
                        type=SystemType.saas,
                        human_readable=connector_template.human_readable,
                        encoded_icon=connector_template.icon,
                        authorization_required=connector_template.authorization_required,
                        user_guide=connector_template.user_guide,
                        supported_actions=connector_template.supported_actions,
                    )
                )

    if (system_type == SystemType.manual or system_type is None) and (
        ActionType.access in action_types or ActionType.erasure in action_types
    ):
        manual_types: list[str] = sorted(
            [
                manual_type.value
                for manual_type in ConnectionType
                if manual_type == ConnectionType.manual_webhook
                and is_match(manual_type.value)
            ]
        )
        connection_system_types.extend(
            [
                ConnectionSystemTypeMap(
                    identifier=item,
                    type=SystemType.manual,
                    human_readable=ConnectionType(item).human_readable,
                    supported_actions=[ActionType.access, ActionType.erasure],
                )
                for item in manual_types
            ]
        )

    if system_type == SystemType.email or system_type is None:
        email_types: list[str] = sorted(
            [
                email_type.value
                for email_type in ConnectionType
                if email_type
                in ERASURE_EMAIL_CONNECTOR_TYPES + CONSENT_EMAIL_CONNECTOR_TYPES
                and is_match(email_type.value)
                and (  # include consent or erasure connectors if requested, respectively
                    (
                        ActionType.consent in action_types
                        and email_type in CONSENT_EMAIL_CONNECTOR_TYPES
                    )
                    or (
                        ActionType.erasure in action_types
                        and email_type in ERASURE_EMAIL_CONNECTOR_TYPES
                    )
                )
            ]
        )
        connection_system_types.extend(
            [
                ConnectionSystemTypeMap(
                    identifier=email_type,
                    type=SystemType.email,
                    human_readable=ConnectionType(email_type).human_readable,
                    supported_actions=[
                        (
                            ActionType.consent
                            if ConnectionType(email_type)
                            in CONSENT_EMAIL_CONNECTOR_TYPES
                            else ActionType.erasure
                        )
                    ],
                )
                for email_type in email_types
            ]
        )

    return connection_system_types



"""Add connection_type to dataset namespace meta

Revision ID: 4ebe0766021b
Revises: 36ad82edb38e
Create Date: 2024-10-16 15:25:08.544946

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4ebe0766021b"
down_revision = "36ad82edb38e"
branch_labels = None
depends_on = None


def upgrade():
    # Populate the connection_type field in the namespace meta for all datasets
    # that have a namespace key in their fides_meta json field.

    # If the namespace dict has a project_id key, we know it's a BigQuery dataset.
    op.execute(
        """
        update ctl_datasets
        set fides_meta = jsonb_set(fides_meta::jsonb, '{namespace, connection_type}', '"bigquery"', true)
        where (fides_meta::jsonb ? 'namespace') and ((fides_meta::jsonb ->> 'namespace')::jsonb ? 'project_id');
        """
    )

    # If the namespace dict has a database_instance_id, we know it's an RDS MySQL dataset.
    op.execute(
        """
        update ctl_datasets
        set fides_meta = jsonb_set(fides_meta::jsonb, '{namespace, connection_type}', '"rds_mysql"', true)
        where (fides_meta::jsonb ? 'namespace') and ((fides_meta::jsonb ->> 'namespace')::jsonb ? 'database_instance_id');
        """
    )


def downgrade():
    # Remove the connection_type field from the namespace meta for all datasets
    # that have a namespace key in their fides_meta json field.
    op.execute(
        """
        update ctl_datasets
        set fides_meta  = fides_meta::jsonb #- '{namespace, connection_type}'
        where fides_meta::jsonb ? 'namespace';
        """
    )



"""update privacy request with source and submitted_by

Revision ID: 896ea3803770
Revises: ffee79245c9a
Create Date: 2024-08-15 23:08:00.169034

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "896ea3803770"
down_revision = "ffee79245c9a"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("consentrequest", sa.Column("source", sa.String(), nullable=True))
    op.add_column(
        "privacyrequest", sa.Column("submitted_by", sa.String(), nullable=True)
    )
    op.add_column("privacyrequest", sa.Column("source", sa.String(), nullable=True))
    op.create_foreign_key(
        "privacyrequest_submitted_by_fkey",
        "privacyrequest",
        "fidesuser",
        ["submitted_by"],
        ["id"],
        ondelete="SET NULL",
    )


def downgrade():
    op.drop_constraint(
        "privacyrequest_submitted_by_fkey", "privacyrequest", type_="foreignkey"
    )
    op.drop_column("privacyrequest", "source")
    op.drop_column("privacyrequest", "submitted_by")
    op.drop_column("consentrequest", "source")



from typing import Optional
from urllib.parse import quote, quote_plus, urlencode

from pydantic import Field, ValidationInfo, field_validator
from pydantic_settings import SettingsConfigDict

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__REDIS__"


class RedisSettings(FidesSettings):
    """Configuration settings for Redis."""

    charset: str = Field(
        default="utf8",
        description="Character set to use for Redis, defaults to 'utf8'. Not recommended to change.",
    )
    db_index: int = Field(
        default=0,
        description="The application will use this index in the Redis cache to cache data.",
    )
    decode_responses: bool = Field(
        default=True,
        description="Whether or not to automatically decode the values fetched from Redis. Decodes using the `charset` configuration value.",
    )
    default_ttl_seconds: int = Field(
        default=604800,
        description="The number of seconds for which data will live in Redis before automatically expiring.",
    )
    enabled: bool = Field(
        default=True,
        description="Whether the application's Redis cache should be enabled. Only set to false for certain narrow uses of the application.",
    )
    host: str = Field(
        default="redis",
        description="The network address for the application Redis cache.",
    )
    identity_verification_code_ttl_seconds: int = Field(
        default=600,
        description="Sets TTL for cached identity verification code as part of subject requests.",
    )
    password: str = Field(
        default="testpassword",
        description="The password with which to login to the Redis cache.",
    )
    port: int = Field(
        default=6379,
        description="The port at which the application cache will be accessible.",
    )
    ssl: bool = Field(
        default=False,
        description="Whether the application's connections to the cache should be encrypted using TLS.",
    )
    ssl_cert_reqs: Optional[str] = Field(
        default="required",
        description="If using TLS encryption, set this to 'required' if you wish to enforce the Redis cache to provide a certificate. Note that not all cache providers support this without setting ssl_ca_certs (e.g. AWS Elasticache).",
    )
    ssl_ca_certs: str = Field(
        default="",
        description="If using TLS encryption rooted with a custom Certificate Authority, set this to the path of the CA certificate.",
    )
    user: str = Field(
        default="", description="The user with which to login to the Redis cache."
    )

    # This relies on other values to get built so must be last
    connection_url: Optional[str] = Field(
        default=None,
        description="A full connection URL to the Redis cache. If not specified, this URL is automatically assembled from the host, port, password and db_index specified above.",
        exclude=True,
    )

    @field_validator("connection_url", mode="before")
    @classmethod
    def assemble_connection_url(
        cls,
        v: Optional[str],
        info: ValidationInfo,
    ) -> str:
        """Join Redis connection credentials into a connection string"""
        if isinstance(v, str):
            # If the whole URL is provided via the config, preference that
            return v

        connection_protocol = "redis"
        params_str = ""
        use_tls = info.data.get("ssl", False)

        # These vars are intentionally fetched with `or ""` as the default to account
        # for the edge case where `None` is explicitly set in `values` by Pydantic because
        # it is not overridden by the config file or an env var
        user = info.data.get("user") or ""
        password = info.data.get("password") or ""
        db_index = info.data.get("db_index") or ""
        if use_tls:
            # If using TLS update the connection URL format
            connection_protocol = "rediss"
            cert_reqs = info.data.get("ssl_cert_reqs", "none")
            params = {"ssl_cert_reqs": quote_plus(cert_reqs)}
            if ssl_ca_certs := info.data.get("ssl_ca_certs", ""):
                params["ssl_ca_certs"] = quote(ssl_ca_certs, safe="/")
            params_str = "?" + urlencode(params, quote_via=quote, safe="/")

        # Configure a basic auth prefix if either user or password is provided, e.g.
        # redis://<user>:<password>@<host>
        auth_prefix = ""
        if password or user:
            auth_prefix = f"{quote_plus(user)}:{quote_plus(password)}@"

        connection_url = f"{connection_protocol}://{auth_prefix}{info.data.get('host', '')}:{info.data.get('port', '')}/{db_index}{params_str}"
        return connection_url

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)



"""Add consent request

Revision ID: 021d288d0ce3
Revises: a0e6feb5bdc8
Create Date: 2022-09-17 01:26:43.187484

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "021d288d0ce3"
down_revision = "a0e6feb5bdc8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "consentrequest",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("provided_identity_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["provided_identity_id"],
            ["providedidentity.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_consentrequest_id"), "consentrequest", ["id"], unique=False
    )
    op.alter_column(
        "consent", "provided_identity_id", existing_type=sa.VARCHAR(), nullable=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "consent", "provided_identity_id", existing_type=sa.VARCHAR(), nullable=True
    )
    op.drop_index(op.f("ix_consentrequest_id"), table_name="consentrequest")
    op.drop_table("consentrequest")
    # ### end Alembic commands ###



"""add_attentive_connector

Revision ID: 54102bba36de
Revises: 50180bbbb959
Create Date: 2023-03-09 22:08:17.480643

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "54102bba36de"
down_revision = "50180bbbb959"
branch_labels = None
depends_on = None


def upgrade():
    # Add 'attentive' to ConnectionType enum and remove 'email'
    op.execute("delete from connectionconfig where connection_type in ('email')")
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', "
        "'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', "
        "'manual_webhook', 'timescale', 'fides', 'sovrn', 'attentive')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")

    # Rename field to something more generic
    op.alter_column(
        "privacyrequest",
        "awaiting_consent_email_send_at",
        new_column_name="awaiting_email_send_at",
    )

    # Rename awaiting_consent_email_send enum value to awaiting_email_send
    op.execute(
        "ALTER TYPE privacyrequeststatus RENAME VALUE 'awaiting_consent_email_send' TO 'awaiting_email_send'"
    )


def downgrade():
    # Remove attentive from the connectiontype enum and restore email
    op.execute("delete from connectionconfig where connection_type in ('attentive')")
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', "
        "'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', "
        "'email', 'manual_webhook', 'timescale', 'fides', 'sovrn')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")

    # Revert field name
    op.alter_column(
        "privacyrequest",
        "awaiting_email_send_at",
        new_column_name="awaiting_consent_email_send_at",
    )

    # Revert awaiting_email_send enum value to awaiting_consent_email_send
    op.execute(
        "ALTER TYPE privacyrequeststatus RENAME VALUE 'awaiting_email_send' TO 'awaiting_consent_email_send'"
    )



import re

from anyascii import anyascii  # type: ignore


def to_snake_case(text: str) -> str:
    """
    Returns a snake-cased str based upon the input text.

    Examples:
        "foo bar" -> "foo_bar"
        "foo\nbar" -> "foo_bar"
        "foo\tbar" -> "foo_bar"
        "foo-bar" -> "foo_bar"
        "foo*bar" -> "foobar"
    """
    text = anyascii(text).lower().strip()
    text = re.sub(r"[^\w\s-]", "", text)
    text = re.sub(r"[\s-]+", "_", text)
    text = re.sub(r"^-+|-+$", "", text)
    return text



from requests import PreparedRequest

from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.schemas.saas.strategy_configuration import (
    BearerAuthenticationConfiguration,
)
from fides.api.service.authentication.authentication_strategy import (
    AuthenticationStrategy,
)
from fides.api.util.saas_util import assign_placeholders


class BearerAuthenticationStrategy(AuthenticationStrategy):
    """
    Replaces the token placeholder with the actual credentials
    and uses it to add a bearer authentication header to the incoming request.
    """

    name = "bearer"
    configuration_model = BearerAuthenticationConfiguration

    def __init__(self, configuration: BearerAuthenticationConfiguration):
        self.token = configuration.token

    def add_authentication(
        self, request: PreparedRequest, connection_config: ConnectionConfig
    ) -> PreparedRequest:
        """Add bearer authentication to the request"""
        request.headers["Authorization"] = "Bearer " + assign_placeholders(  # type: ignore
            self.token, connection_config.secrets  # type: ignore
        )
        return request



"""Add Attachment and AttachmentReference tables

Revision ID: bd875a8b5d96
Revises: 82883b0df5e4
Create Date: 2025-02-19 14:09:07.124680

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "bd875a8b5d96"
down_revision = "82883b0df5e4"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # create attachment table
    op.create_table(
        "attachment",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("user_id", sa.String(), nullable=True),
        sa.Column("file_name", sa.String(), nullable=False),
        sa.Column("attachment_type", sa.String(), nullable=False),
        sa.Column("storage_key", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["fidesuser.id"],
            ondelete="SET NULL",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_attachment_id"), "attachment", ["id"], unique=False)

    # create attachment_reference table
    op.create_table(
        "attachment_reference",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("attachment_id", sa.String(), nullable=False),
        sa.Column("reference_id", sa.String(), nullable=False),
        sa.Column("reference_type", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["attachment_id"],
            ["attachment.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_attachment_reference_id"),
        "attachment_reference",
        ["id"],
        unique=False,
    )
    op.create_unique_constraint(
        "_attachment_reference_uc",
        "attachment_reference",
        ["attachment_id", "reference_id"],
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_attachment_reference_id"), table_name="attachment_reference")
    op.drop_table("attachment_reference")
    op.drop_index(op.f("ix_attachment_id"), table_name="attachment")
    op.drop_table("attachment")
    # ### end Alembic commands ###



from datetime import datetime
from typing import Any, Dict, Union

from bson import ObjectId
from pydantic import ValidationError

from fides.api.schemas.storage.storage import (
    SUPPORTED_STORAGE_SECRETS,
    StorageSecretsGCS,
    StorageSecretsS3,
    StorageType,
)
from fides.api.schemas.storage.storage_secrets_docs_only import possible_storage_secrets


def get_schema_for_secrets(
    storage_type: Union[StorageType, str],
    secrets: possible_storage_secrets,
) -> SUPPORTED_STORAGE_SECRETS:
    """
    Returns the secrets that pertain to `storage_type` represented as a Pydantic schema
    for validation purposes.
    """
    if not isinstance(storage_type, StorageType):
        # try to coerce into an enum
        try:
            storage_type = StorageType[storage_type]
        except KeyError:
            raise ValueError(
                "storage_type argument must be a valid StorageType enum member."
            )
    try:
        schema = {
            StorageType.s3: StorageSecretsS3,
            StorageType.gcs: StorageSecretsGCS,
        }[storage_type]
    except KeyError:
        raise ValueError(
            f"`storage_type` {storage_type} has no supported `secrets` validation."
        )

    try:
        return schema.model_validate(secrets)  # type: ignore
    except ValidationError as exc:
        # Pydantic requires validators raise either a ValueError, TypeError, or AssertionError
        # so this exception is cast into a `ValueError`.
        errors = [f"{err['msg']} {str(err['loc'])}" for err in exc.errors()]
        raise ValueError(errors)


def storage_json_encoder(field: Any) -> Union[str, Dict[str, str]]:
    """Specify str format for datetime objects"""
    if isinstance(field, datetime):
        return field.strftime("%Y-%m-%dT%H:%M:%S")
    if isinstance(field, ObjectId):
        return {"$oid": str(field)}
    return field



"""fix message template indexes

Revision ID: 52b2cfd1af47
Revises: f17f92237383
Create Date: 2023-09-15 21:29:23.060892

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "52b2cfd1af47"
down_revision = "f17f92237383"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index("ix_messagingtemplate_id", table_name="messaging_template")
    op.drop_index("ix_messagingtemplate_key", table_name="messaging_template")
    op.create_index(
        op.f("ix_messaging_template_id"), "messaging_template", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_messaging_template_key"), "messaging_template", ["key"], unique=True
    )


def downgrade():
    op.drop_index(op.f("ix_messaging_template_key"), table_name="messaging_template")
    op.drop_index(op.f("ix_messaging_template_id"), table_name="messaging_template")
    op.create_index(
        "ix_messagingtemplate_key", "messaging_template", ["key"], unique=False
    )
    op.create_index(
        "ix_messagingtemplate_id", "messaging_template", ["id"], unique=False
    )



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class MaskingConfiguration(BaseModel):
    """Base class for masking configuration"""


class FormatPreservationConfig(BaseModel):
    """option to preserve format in masking"""

    suffix: str


class StringRewriteMaskingConfiguration(MaskingConfiguration):
    """Configuration for the StringRewriteMaskingStrategy"""

    rewrite_value: str
    format_preservation: Optional[FormatPreservationConfig] = None


class HashMaskingConfiguration(MaskingConfiguration):
    """Configuration for the HashMaskingStrategy"""

    class Algorithm(Enum):
        """Denotes hashing algorithms supported by HashMaskingStrategy"""

        SHA_256 = "SHA-256"
        SHA_512 = "SHA-512"

    algorithm: Algorithm = Algorithm.SHA_256
    format_preservation: Optional[FormatPreservationConfig] = None


class AesEncryptionMaskingConfiguration(MaskingConfiguration):
    """Configuration for the AES encryption masking strategy"""

    class Mode(Enum):
        GCM = "GCM"

    mode: Mode = Mode.GCM
    format_preservation: Optional[FormatPreservationConfig] = None


class RandomStringMaskingConfiguration(MaskingConfiguration):
    """Configuration for the RandomStringRewriteMaskingStrategy"""

    length: int = 30
    format_preservation: Optional[FormatPreservationConfig] = None


class HmacMaskingConfiguration(MaskingConfiguration):
    """Configuration for the HmacMaskingStrategy"""

    class Algorithm(Enum):
        """Denotes hashing algorithms supported by HmacMaskingStrategy
        If none supplied, SHA-256 used by default.
        """

        sha_256 = "SHA-256"
        sha_512 = "SHA-512"

    algorithm: Algorithm = Algorithm.sha_256
    format_preservation: Optional[FormatPreservationConfig] = None


class NullMaskingConfiguration(MaskingConfiguration):
    """Configuration for the NullMaskingStrategy - this is the simplest masking strategy, no details are supplied"""



from typing import Any

from cassandra.util import Date, Duration, OrderedMapSerializedKey, SortedSet, Time


def scylla_to_native_python(data: Any) -> Any:  # type: ignore[misc] # pylint: disable=too-many-return-statements
    """Recursively convert some non-standard Scylla types to native Python types

    Scylla also supports 3 kinds of collections: maps, lists, and sets
    """
    if isinstance(data, OrderedMapSerializedKey):
        return scylla_to_native_python(dict(data))
    if isinstance(data, SortedSet):
        return scylla_to_native_python(set(data))
    if isinstance(data, Date):
        return data.date()
    if isinstance(data, Time):
        return data.time()
    if isinstance(data, Duration):
        # This is in a tuple format of months, days, nanoseconds
        # For now, returning as-is
        return data
    if isinstance(data, list):
        return [scylla_to_native_python(elem) for elem in data]
    if isinstance(data, set):
        return {scylla_to_native_python(elem) for elem in data}
    if isinstance(data, dict):
        return {
            scylla_to_native_python(key): scylla_to_native_python(value)
            for key, value in data.items()
        }

    return data



"""property specific messaging migration data


Revision ID: 5f96f13be8ec
Revises: 2736c942faa2
Create Date: 2024-05-28 14:52:08.114674

"""

import random
import string
import uuid
from typing import Any, Dict, List

import sqlalchemy as sa
from alembic import op
from sqlalchemy import bindparam, text

# revision identifiers, used by Alembic.
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.engine import Connection, LegacyCursorResult
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql.elements import TextClause

from fides.api.alembic.migrations.helpers.database_functions import generate_record_id
from fides.api.schemas.messaging.messaging import MessagingActionType

# revision identifiers, used by Alembic.
revision = "5f96f13be8ec"
down_revision = "2736c942faa2"
branch_labels = None
depends_on = None


DEFAULT_MESSAGING_TEMPLATES: Dict[str, Any] = {
    MessagingActionType.SUBJECT_IDENTITY_VERIFICATION.value: {
        "label": "Subject identity verification",
        "content": {
            "subject": "Your one-time code is {{code}}",
            "body": "Your privacy request verification code is {{code}}. Please return to the Privacy Center and enter the code to continue. This code will expire in {{minutes}} minutes.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_RECEIPT.value: {
        "label": "Privacy request received",
        "content": {
            "subject": "Your privacy request has been received",
            "body": "Your privacy request has been received. We will get back to you shortly.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE.value: {
        "label": "Privacy request approved",
        "content": {
            "subject": "Your privacy request has been approved",
            "body": "Your privacy request has been approved and is currently processing.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY.value: {
        "label": "Privacy request denied",
        "content": {
            "subject": "Your privacy request has been denied",
            "body": "Your privacy request has been denied. {{denial_reason}}.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS.value: {
        "label": "Access request completed",
        "content": {
            "subject": "Your data is ready to be downloaded",
            "body": "Your access request has been completed and can be downloaded at {{download_link}}. For security purposes, this secret link will expire in {{days}} days.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION.value: {
        "label": "Erasure request completed",
        "content": {
            "subject": "Your data has been deleted",
            "body": "Your erasure request has been completed.",
        },
    },
}

AUTO_MIGRATED_STRING = "auto-migrated"


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind: Connection = op.get_bind()

    # STEP 1: Ensure we have exactly 1 default property in the DB
    existing_properties: LegacyCursorResult = bind.execute(
        text("select id from plus_property;")
    )
    if len(existing_properties.all()) == 1:
        # If exactly one property exists, assume this is the "default"
        bind.execute(text("UPDATE plus_property SET is_default = TRUE;"))
        only_property: LegacyCursorResult = bind.execute(
            text("select id from plus_property;")
        ).first()
        default_property_id = only_property["id"]
    else:
        # Create new property, label as default
        insert_into_property_query: TextClause = text(
            "INSERT INTO plus_property (id, name, is_default, type)"
            "VALUES (:id, :name, :is_default, :type)"
        )
        characters = string.ascii_uppercase + string.digits
        new_property_id: str = "FDS-" + "".join(random.choices(characters, k=6))
        default_property_id = new_property_id

        new_property: Dict[str, Any] = {
            "id": new_property_id,
            "name": "Default Property Autogenerated",
            "is_default": True,
            "type": "website",
        }
        try:
            bind.execute(insert_into_property_query, new_property)
        except IntegrityError as exc:
            raise Exception(
                f"Fides attempted to create a new default property but got error: {exc}. "
            )

    # STEP 2: Ensure all default templates are saved to the DB
    existing_templates: LegacyCursorResult = bind.execute(
        text("SELECT id, type from messaging_template;")
    )
    templates_from_db: List[str] = [
        template.type for template in existing_templates.all()
    ]
    for template_type, template in DEFAULT_MESSAGING_TEMPLATES.items():
        # if a template is not already in the DB, save it with defaults
        if template_type not in templates_from_db:
            insert_into_messaging_template_query: TextClause = text(
                "INSERT INTO messaging_template (id, type, content, is_enabled)"
                "VALUES (:id, :type, :content, :is_enabled)"
            )
            new_messaging_template: Dict[str, Any] = {
                "id": generate_record_id("autogenerated-mes"),
                "type": template_type,
                "content": template["content"],
                "is_enabled": False,
            }
            try:
                bind.execute(
                    insert_into_messaging_template_query.bindparams(
                        bindparam("content", type_=JSONB)
                    ),
                    new_messaging_template,
                )
            except IntegrityError as exc:
                raise Exception(
                    f"Fides attempted to create a new messaging_template but got error: {exc}. "
                )

    # STEP 3: Ensure all saved messaging templates are linked to the default property
    updated_templates: LegacyCursorResult = bind.execute(
        text("select id, type from messaging_template;")
    )
    for template in updated_templates:
        insert_into_messaging_template_to_property_query: TextClause = text(
            "INSERT INTO messaging_template_to_property (id, messaging_template_id, property_id)"
            "VALUES (:id, :messaging_template_id, :property_id)"
        )
        new_messaging_template_to_property: Dict[str, Any] = {
            "id": generate_record_id("mes"),
            "messaging_template_id": template["id"],
            "property_id": default_property_id,
        }
        bind.execute(
            insert_into_messaging_template_to_property_query,
            new_messaging_template_to_property,
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # Reverse data migration: remove templates that were automatically created by the forward migration
    bind = op.get_bind()
    bind.execute(text("DELETE FROM messaging_template_to_property;"))
    bind.execute(text("DELETE FROM messaging_template WHERE id LIKE '%autogenerated%'"))
    bind.execute(
        text(
            "DELETE FROM plus_property WHERE name LIKE 'Default Property Autogenerated'"
        )
    )
    # ### end Alembic commands ###



from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence

from fideslang.models import PrivacyDeclaration, System
from pydantic import ConfigDict, Field
from pydantic.main import BaseModel

from fides.api.schemas.connection_configuration.connection_config import (
    ConnectionConfigurationResponse,
)
from fides.api.schemas.user import UserResponse


class PrivacyDeclarationResponse(PrivacyDeclaration):
    """Extension of base pydantic model to include DB `id` field in the response"""

    id: str = Field(
        description="The database-assigned ID of the privacy declaration on the system. This is meant to be a read-only field, returned only in API responses"
    )


class BasicSystemResponse(System):
    """
    Extension of base pydantic model to include additional fields on the DB model that
    are relevant in API responses.

    This is still meant to be a "lightweight" model that does not reference relationships
    that may require additional querying beyond the `System` db table.
    """

    created_at: datetime


class SystemResponse(BasicSystemResponse):
    """Extension of base pydantic response model to include additional relationship fields that
    may require extra DB queries, like `privacy_declarations`, connection_config fields and cookies.

    This response model is generally useful for an API that returns a detailed view of _single_
    System record. Attempting to return bulk results with this model can lead to n+1 query issues.
    """

    privacy_declarations: Sequence[PrivacyDeclarationResponse] = Field(  # type: ignore[assignment]
        description=PrivacyDeclarationResponse.__doc__,
    )

    connection_configs: Optional[ConnectionConfigurationResponse] = Field(
        description=ConnectionConfigurationResponse.__doc__,
    )

    data_stewards: Optional[List[UserResponse]] = Field(
        description="System managers of the current system",
    )


class SystemHistoryResponse(BaseModel):
    """Response schema for a single system history entry"""

    edited_by: Optional[str] = None
    system_id: str
    before: Dict[str, Any]
    after: Dict[str, Any]
    created_at: datetime
    model_config = ConfigDict(from_attributes=True)



from loguru import logger

from fides.api.common_exceptions import TraversalError, ValidationError
from fides.api.graph.graph import DatasetGraph
from fides.api.graph.traversal import Traversal
from fides.api.models.connectionconfig import ConnectionType
from fides.api.models.datasetconfig import convert_dataset_to_graph
from fides.api.schemas.dataset import DatasetTraversalDetails
from fides.api.util.saas_util import merge_datasets
from fides.service.dataset.dataset_validator import (
    DatasetValidationContext,
    DatasetValidationStep,
)


class TraversalValidationStep(DatasetValidationStep):
    """Validates dataset traversability"""

    def validate(self, context: DatasetValidationContext) -> None:
        if not context.connection_config:
            logger.warning(
                "Skipping traversal validation, no connection config provided"
            )
            return

        try:
            graph = convert_dataset_to_graph(
                context.dataset, context.connection_config.key
            )

            if context.connection_config.connection_type == ConnectionType.saas:
                graph = merge_datasets(
                    graph,
                    context.connection_config.get_saas_config().get_graph(
                        context.connection_config.secrets
                    ),
                )

            complete_graph = DatasetGraph(graph)
            unique_identities = set(complete_graph.identity_keys.values())
            Traversal(complete_graph, {k: None for k in unique_identities})

            context.traversal_details = DatasetTraversalDetails(
                is_traversable=True, msg=None
            )

        except (TraversalError, ValidationError) as err:
            context.traversal_details = DatasetTraversalDetails(
                is_traversable=False, msg=str(err)
            )
